Package: multibias
Type: Package
Title: Multiple Bias Analysis in Causal Inference
Version: 1.7.3
Date: 2026-04-20
Authors@R: person("Paul", "Brendel", email = "pcbrendel@gmail.com",
    role = c("aut", "cre", "cph"))
Maintainer: Paul Brendel <pcbrendel@gmail.com>
Description: Quantify exposure-outcome causal effects
    with adjustment for multiple biases. The functions can
    simultaneously adjust for any combination of uncontrolled confounding,
    exposure/outcome misclassification, and selection bias.
    The underlying method generalizes the combination of inverse
    probability of selection weighting with predictive value weighting.
    Simultaneous multi-bias analysis can be used to enhance the validity
    and transparency of real-world evidence obtained from observational,
    longitudinal studies. Based on the work from Paul Brendel, Aracelis Torres, and Onyebuchi Arah (2023) <doi:10.1093/ije/dyad001>.
License: MIT + file LICENSE
Encoding: UTF-8
LazyData: true
Depends: R (>= 4.2.0)
RoxygenNote: 7.3.2
Imports: dplyr (>= 1.1.3), lifecycle (>= 1.0.3), magrittr (>= 2.0.3),
        rlang (>= 1.1.1), broom (>= 1.0.5), purrr (>= 1.0.0), ggplot2
        (>= 3.5.0)
Suggests: knitr, rmarkdown, MASS, testthat (>= 3.0.0), vdiffr (>=
        1.0.5)
URL: https://github.com/pcbrendel/multibias,
        http://www.paulbrendel.com/multibias/
BugReports: https://github.com/pcbrendel/multibias/issues
Config/testthat/edition: 3
VignetteBuilder: knitr
NeedsCompilation: no
Packaged: 2026-04-20 22:45:03 UTC; pbrendel
Author: Paul Brendel [aut, cre, cph]
Repository: CRAN
Date/Publication: 2026-04-20 23:30:02 UTC
Built: R 4.5.2; ; 2026-04-21 01:25:54 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(print,bias_params)
S3method(print,data_observed)
S3method(print,data_validation)
S3method(summary,data_observed)
export(bias_params)
export(data_observed)
export(data_validation)
export(multibias_adjust)
export(multibias_plot)
import(dplyr)
importFrom(broom,tidy)
importFrom(lifecycle,deprecated)
importFrom(magrittr,"%>%")
importFrom(rlang,.data)
importFrom(stats,as.formula)
importFrom(stats,binomial)
importFrom(stats,coef)
importFrom(stats,glm)
importFrom(stats,lm)
importFrom(stats,median)
importFrom(stats,plogis)
importFrom(stats,qnorm)
importFrom(stats,quantile)
importFrom(stats,rbinom)
importFrom(stats,rnorm)
importFrom(stats,sd)
importFrom(stats,vcov)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# multibias <img src="man/figures/logo.png" align="right" />

<!-- badges: start -->

[![R-CMD-check](https://github.com/pcbrendel/multibias/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/pcbrendel/multibias/actions/workflows/R-CMD-check.yaml)
[![cranlogs](https://cranlogs.r-pkg.org/badges/multibias)](https://cran.r-project.org/package=multibias)
<!-- badges: end -->

## Overview

The multibias package is used to adjust for multiple biases in causal
inference when working with observational data. Bias here refers to the
case when the associational estimate of effect does not equal the causal
estimate of effect:

$$(P(Y=1|X=1,C=0) / P(Y=1|X=0,C=0)) \neq (P(Y^{X=1}=1) / P(Y^{X=0}=1))$$

The `multibias_adjust()` function outputs odds ratio estimates adjusted
for any combination of: uncontrolled confounding (**uc**), exposure
misclassification (**em**), outcome misclassification (**om**), and
selection bias (**sel**).

The package also includes several dataframes that are useful for
validating the bias adjustment methods. Each dataframe contains
different combinations of bias as identified by the same prefixing
system. For each bias combination, there is a dataframe with incomplete
information (as would be encountered in the real world) (e.g., `df_uc`)
and a dataframe with complete information that was used to derive the
biased data (e.g., `df_uc_source`).

## Installation

``` r
# install from CRAN
install.packages("multibias")

# install from github using devtools
# library("devtools")
devtools::install_github("pcbrendel/multibias")
```

## Getting started

1.  Represent the observed causal data as a `data_observed` object. Here
    you provide the data, specify the key variables, and list the biases
    present in the data. See list below for the different bias
    combinations that multibias can handle.
2.  Obtain one of the two sources for bias adjustment:
    1.  Bias parameters - via the `bias_params` object. Values for these
        parameters could come from the literature, validation data, or
        expert opinion. Each parameter can be represented as a single
        value or as a probability distribution. See the `bias_params`
        documentation for the full bias models.
    2.  Validation dataframe - via the `data_validation` object. The
        purpose of validation data is to use an external data source to
        transport the necessary causal relationships that are missing in
        the observed data.
3.  Run `multibias_adjust()` using the above inputs to obtain the
    bias-adjusted exposure-outcome odds ratio and confidence interval.
4.  Visualize a Forest Plot of the observed effect estimate against
    various bias-adjusted estimates via `multibias_plot()`.

### Possible bias adjustments

<img src="man/figures/bias_combos_table.png" align="center" width="80%" />

## References

- Brendel PB, Torres AZ, Arah OA, Simultaneous adjustment of
  uncontrolled confounding, selection bias and misclassification in
  multiple-bias modelling, *International Journal of Epidemiology*,
  Volume 52, Issue 4, Pages 1220–1230.
  <https://doi.org/10.1093/ije/dyad001>
- Wood CJC, Tilling K, Bartlett JW, Hughes RA. Quantitative bias
  analysis for mismeasured variables in health research: a review of
  software tools. *BMC Med Res Methodol.* 2025 Aug 1;25(1):187.
  [https://doi.org/10.1186/s12874-025-02635-w](https://pmc.ncbi.nlm.nih.gov/articles/PMC12317562/)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup, message = FALSE---------------------------------------------------
library(multibias)
library(dplyr)

## -----------------------------------------------------------------------------
nhanes <- read.csv("nhanes.csv")

## -----------------------------------------------------------------------------
nhanes_filtered <- nhanes |>
  mutate(alcohol_extreme = case_when(
    alcohol_day_total > 14 * 1.5 & gender_female == 1 ~ 1,
    alcohol_day_total > 28 * 1.5 & gender_female == 0 ~ 1,
    TRUE ~ 0
  )) |>
  filter(age >= 18) |>
  filter(eligstat == 1) |>
  filter(alcohol_12mo > 0)

## -----------------------------------------------------------------------------
exposure_outcome_table <- nhanes_filtered |>
  group_by(alcohol_extreme, mortstat) |>
  summarize(count = n()) |>
  ungroup() |>
  mutate(proportion = count / sum(count))

# Distribution of Exposure and Outcome:
print(exposure_outcome_table)

demographic_table <- nhanes_filtered |>
  group_by(alcohol_extreme) |>
  summarize(
    n = n(),
    age_mean = mean(age),
    age_sd = sd(age),
    gender_prop = mean(gender_female)
  ) |>
  mutate(
    alcohol_extreme = if_else(
      alcohol_extreme == 1, "Extreme", "Non-extreme"
    )
  )

# Demographic Characteristics by Exposure Status:
print(demographic_table)

## -----------------------------------------------------------------------------
# Create observed data object with uncontrolled confounding bias
df_obs1 <- data_observed(
  data = nhanes_filtered,
  bias = "uc",
  exposure = "alcohol_extreme",
  outcome = "mortstat",
  confounders = c("age", "gender_female")
)

summary(df_obs1)

## -----------------------------------------------------------------------------
# Create validation data with smoking information
df_temp1 <- nhanes_filtered |>
  filter(!is.na(smoked_100cigs))

df_val1 <- data_validation(
  data = df_temp1,
  true_exposure = "alcohol_extreme",
  true_outcome = "mortstat",
  confounders = c("age", "gender_female", "smoked_100cigs")
)

# Perform bias adjustment
set.seed(1234)
uc_adjusted <- multibias_adjust(
  df_obs1,
  df_val1,
  bootstrap = TRUE,
  bootstrap_reps = 100
)

# Uncontrolled Confounding Adjusted Results:
print(uc_adjusted)

## -----------------------------------------------------------------------------
# Create observed data object with both biases
df_obs2 <- data_observed(
  data = nhanes_filtered,
  bias = c("em", "uc"),
  exposure = "alcohol_extreme",
  outcome = "mortstat",
  confounders = c("age", "gender_female")
)

# Create validation data with adjusted alcohol consumption
df_temp2 <- nhanes_filtered |>
  filter(!is.na(smoked_100cigs)) |>
  mutate(
    # Assume 50% higher consumption for high SES individuals
    alcohol_adj = if_else(
      income == "$100,000 and Over" | education == "College graduate or above",
      alcohol_day_total * 1.5,
      alcohol_day_total
    )
  ) |>
  mutate(alcohol_extreme_adj = case_when(
    alcohol_adj > 14 * 1.5 & gender_female == 1 ~ 1,
    alcohol_adj > 28 * 1.5 & gender_female == 0 ~ 1,
    TRUE ~ 0
  ))

df_val2 <- data_validation(
  data = df_temp2,
  true_exposure = "alcohol_extreme_adj",
  true_outcome = "mortstat",
  confounders = c("age", "gender_female", "smoked_100cigs"),
  misclassified_exposure = "alcohol_extreme"
)

# Perform bias adjustment
set.seed(1234)
uc_em_adjusted <- multibias_adjust(
  df_obs2,
  df_val2,
  bootstrap = TRUE,
  bootstrap_reps = 100
)

# Exposure Misclassification & Confounding Adjusted Results:
print(uc_em_adjusted)

## -----------------------------------------------------------------------------
# Create observed data object with all three biases
df_obs3 <- data_observed(
  data = nhanes_filtered,
  bias = c("em", "uc", "sel"),
  exposure = "alcohol_extreme",
  outcome = "mortstat",
  confounders = c("age", "gender_female")
)

# Prepare data for selection bias adjustment
df_temp3a <- nhanes_filtered |>
  filter(!is.na(smoked_100cigs)) |>
  mutate(
    alcohol_adj = if_else(
      income == "$100,000 and Over" | education == "College graduate or above",
      alcohol_day_total * 1.5,
      alcohol_day_total
    ),
    alcohol_extreme_adj = case_when(
      alcohol_adj > 14 * 1.5 & gender_female == 1 ~ 1,
      alcohol_adj > 28 * 1.5 & gender_female == 0 ~ 1,
      TRUE ~ 0
    ),
    # Combine survey weights
    weight = if_else(weight_day2 == 0, weight_day1, weight_day2)
  )

# Create selection indicator based on sampling weights
set.seed(1234)
selected_sample <- sample(
  x = df_temp3a$seqn,
  size = nrow(df_temp3a),
  replace = TRUE,
  prob = df_temp3a$weight
)

df_selected_sample <- data.frame()
for (id in selected_sample) {
  df_selected_sample <- rbind(
    df_selected_sample,
    df_temp3a[df_temp3a$seqn == id, ]
  )
}

df_selected_sample$selection <- 1
df_not_selected_sample <- df_temp3a[!(df_temp3a$seqn %in% selected_sample), ]
df_not_selected_sample$selection <- 0

df_temp3b <- rbind(df_selected_sample, df_not_selected_sample)

# Create validation data with selection information
df_val3 <- data_validation(
  data = df_temp3b,
  true_exposure = "alcohol_extreme_adj",
  true_outcome = "mortstat",
  confounders = c("age", "gender_female", "smoked_100cigs"),
  misclassified_exposure = "alcohol_extreme",
  selection = "selection"
)

# Perform final bias adjustment
set.seed(1234)
final_adjusted <- multibias_adjust(
  df_obs3,
  df_val3,
  bootstrap = TRUE,
  bootstrap_reps = 100
)

# Triple Bias Adjusted Results:
print(final_adjusted)

## -----------------------------------------------------------------------------
multibias_plot(
  df_obs1,
  multibias_result_list = list(
    "Adjusted - Single Bias" = uc_adjusted,
    "Adjusted - Double Biases" = uc_em_adjusted,
    "Adjusted - Triple Biases" = final_adjusted
  ),
  log_scale = TRUE
)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup--------------------------------------------------------------------
library(multibias)

## ----out.width = '70%', echo = FALSE------------------------------------------
knitr::include_graphics("img/uc_emc_sel_DAG.png")

## ----eval = TRUE--------------------------------------------------------------
head(df_uc_em_sel)

## ----eval = TRUE--------------------------------------------------------------
df_observed <- data_observed(
  data = df_uc_em_sel,
  bias = c("uc", "em", "sel"),
  exposure = "Xstar",
  outcome = "Y",
  confounders = c("C1", "C2", "C3")
)

print(df_observed)
summary(df_observed)

## ----eval = TRUE--------------------------------------------------------------
u_model <- glm(U ~ X + Y,
  family = binomial(link = "logit"),
  data = df_uc_em_sel_source
)
x_model <- glm(X ~ Xstar + Y + C1 + C2 + C3,
  family = binomial(link = "logit"),
  data = df_uc_em_sel_source
)
s_model <- glm(S ~ Xstar + Y + C1 + C2 + C3,
  family = binomial(link = "logit"),
  data = df_uc_em_sel_source
)

## -----------------------------------------------------------------------------
bp <- bias_params(
  coef_list = list(
    u = c(
      rnorm(1, mean = u_model$coef[1], sd = summary(u_model)$coef[1, 2]),
      rnorm(1, mean = u_model$coef[2], sd = summary(u_model)$coef[2, 2]),
      rnorm(1, mean = u_model$coef[3], sd = summary(u_model)$coef[3, 2])
    ),
    x = c(
      rnorm(1, mean = x_model$coef[1], sd = summary(x_model)$coef[1, 2]),
      rnorm(1, mean = x_model$coef[2], sd = summary(x_model)$coef[2, 2]),
      rnorm(1, mean = x_model$coef[3], sd = summary(x_model)$coef[3, 2]),
      rnorm(1, mean = x_model$coef[4], sd = summary(x_model)$coef[4, 2]),
      rnorm(1, mean = x_model$coef[5], sd = summary(x_model)$coef[5, 2]),
      rnorm(1, mean = x_model$coef[6], sd = summary(x_model)$coef[6, 2])
    ),
    s = c(
      rnorm(1, mean = s_model$coef[1], sd = summary(s_model)$coef[1, 2]),
      rnorm(1, mean = s_model$coef[2], sd = summary(s_model)$coef[2, 2]),
      rnorm(1, mean = s_model$coef[3], sd = summary(s_model)$coef[3, 2]),
      rnorm(1, mean = s_model$coef[4], sd = summary(s_model)$coef[4, 2]),
      rnorm(1, mean = s_model$coef[5], sd = summary(s_model)$coef[5, 2]),
      rnorm(1, mean = s_model$coef[6], sd = summary(s_model)$coef[6, 2])
    )
  )
)

## -----------------------------------------------------------------------------
multibias_adjust(
  data_observed = df_observed,
  bias_params = bp,
  bootstrap = TRUE,
  bootstrap_reps = 10,
  level = 0.95
)

# Adjust for uncontrolled confounding and outcome misclassification

# the following functions feed into adjust_uc_om():
# adjust_uc_om_val() (data_validation input, method: imputation),
# adjust_uc_om_coef_single() (bias_params input, method: imputation),
# adjust_uc_om_coef_multinom() (bias_params input, method: weighting)

adjust_uc_om_val <- function(
    data_observed,
    data_validation) {
  if (!all(data_observed$confounders %in% data_validation$confounders)) {
    stop(
      "All confounders in observed data must be present in validation data.",
      call. = FALSE
    )
  }

  if (
    (length(data_validation$confounders) - length(data_observed$confounders) != 1) ||
      (is.null(data_validation$misclassified_outcome))
  ) {
    stop(
      paste0(
        "Attempting to adjust for unobserved confounding from one confounder and outcome misclassification.",
        "\n",
        "Validation data must have: 1) one more confounder than the observed data, 2) a true and misclassified outcome specified."
      ),
      call. = FALSE
    )
  }

  n <- nrow(data_observed$data)

  df <- data.frame(
    X = data_observed$data[, data_observed$exposure],
    Ystar = data_observed$data[, data_observed$outcome]
  )
  df <- bind_cols(
    df,
    data_observed$data %>%
      select(all_of(data_observed$confounders))
  )

  df_val <- data.frame(
    X = data_validation$data[, data_validation$true_exposure],
    Y = data_validation$data[, data_validation$true_outcome],
    Ystar = data_validation$data[, data_validation$misclassified_outcome]
  )

  uc <- setdiff(data_validation$confounders, data_observed$confounders)
  df_val$U <- data_validation$data[, uc]
  df_val <- bind_cols(
    df_val,
    data_validation$data %>%
      select(all_of(data_observed$confounders))
  )

  force_match(
    df$X,
    df_val$X,
    "Outcomes from both datasets must both be binary or both be continuous."
  )
  force_binary(
    df_val$U,
    "Uncontrolled confounder in validation data must be a binary integer."
  )
  force_binary(
    df$Ystar,
    "Outcome in observed data must be a binary integer."
  )
  force_binary(
    df_val$Ystar,
    "Misclassified outcome in validation data must be a binary integer."
  )
  force_binary(
    df_val$Y,
    "True outcome in validation data must be a binary integer."
  )

  y_mod <- glm(Y ~ X + Ystar + . - U,
    family = binomial(link = "logit"),
    data = df_val
  )

  y_mod_coefs <- coef(y_mod)
  y_mod_se <- sqrt(diag(vcov(y_mod)))

  # Sample coefficients independently using their standard errors
  y_mod_coefs_sampled <- rnorm(length(y_mod_coefs), y_mod_coefs, y_mod_se)
  y_pred <- y_mod_coefs_sampled[1]

  for (i in 2:length(y_mod_coefs_sampled)) {
    var_name <- names(y_mod_coefs)[i]
    y_pred <- y_pred + df[[var_name]] * y_mod_coefs_sampled[i]
  }

  df$Ypred <- rbinom(n, 1, plogis(y_pred))

  u_mod <- glm(U ~ X + Y,
    family = binomial(link = "logit"),
    data = df_val
  )

  u_mod_coefs <- coef(u_mod)
  u_mod_se <- sqrt(diag(vcov(u_mod)))

  # Sample coefficients independently using their standard errors
  u_mod_coefs_sampled <- rnorm(length(u_mod_coefs), u_mod_coefs, u_mod_se)
  u_pred <- u_mod_coefs_sampled[1]

  for (i in 2:length(u_mod_coefs_sampled)) {
    var_name <- names(u_mod_coefs)[i]
    var_name <- gsub("Y", "Ypred", var_name) # col Y is not in df
    u_pred <- u_pred + df[[var_name]] * u_mod_coefs_sampled[i]
  }

  df$Upred <- rbinom(n, 1, plogis(u_pred))

  final <- glm(
    Ypred ~ X + Upred + . - Ystar,
    family = binomial(link = "logit"),
    data = df
  )

  return(final)
}


adjust_uc_om_coef_single <- function(
    data_observed,
    u_model_coefs,
    y_model_coefs) {
  data <- data_observed$data
  n <- nrow(data)
  confounders <- data_observed$confounders
  len_c <- length(confounders)
  len_u_coefs <- length(u_model_coefs)
  len_y_coefs <- length(y_model_coefs)

  x <- data[, data_observed$exposure]
  ystar <- data[, data_observed$outcome]

  force_binary(ystar, "Outcome must be a binary integer.")
  force_len(
    len_u_coefs,
    3,
    paste0(
      "Incorrect length of U model coefficients. ",
      "Length should equal 3."
    )
  )
  force_len(
    len_y_coefs,
    3 + len_c,
    paste0(
      "Incorrect length of Y model coefficients. ",
      "Length should equal 3 + number of confounders."
    )
  )

  # Extract model coefficients
  u1_0 <- u_model_coefs[1]
  u1_x <- u_model_coefs[2]
  u1_y <- u_model_coefs[3]

  y1_0 <- y_model_coefs[1]
  y1_x <- y_model_coefs[2]
  y1_ystar <- y_model_coefs[3]
  y1_c <- y_model_coefs[4:len_y_coefs]

  # Create base dataframe
  df <- data.frame(X = x, Ystar = ystar)

  # Add confounders if they exist
  if (!is.null(confounders)) {
    for (i in seq_along(confounders)) {
      df[[paste0("C", i)]] <- data[, confounders[i]]
    }
  }

  # Calculate Y predictions
  y_lp <- y1_0 + y1_x * df$X + y1_ystar * df$Ystar
  if (!is.null(confounders)) {
    C_matrix <- as.matrix(
      df[, paste0("C", seq_along(confounders)), drop = FALSE]
    )
    y_lp <- y_lp + C_matrix %*% y1_c
  }
  df$Ypred <- rbinom(n, 1, plogis(y_lp))

  # Calculate U predictions
  df$Upred <- rbinom(n, 1, plogis(u1_0 + u1_x * df$X + u1_y * df$Ypred))

  # Construct final model formula
  model_terms <- c("X", "Upred")
  if (!is.null(confounders)) {
    model_terms <- c(model_terms, paste0("C", seq_along(confounders)))
  }
  model_formula <- as.formula(
    paste("Ypred ~", paste(model_terms, collapse = " + "))
  )

  # Fit final model
  final <- glm(
    model_formula,
    family = binomial(link = "logit"),
    data = df
  )

  return(final)
}


adjust_uc_om_coef_multinom <- function(
    data_observed,
    u1y0_model_coefs,
    u0y1_model_coefs,
    u1y1_model_coefs) {
  data <- data_observed$data
  n <- nrow(data)
  confounders <- data_observed$confounders
  len_c <- length(confounders)
  len_u1y0_coefs <- length(u1y0_model_coefs)
  len_u0y1_coefs <- length(u0y1_model_coefs)
  len_u1y1_coefs <- length(u1y1_model_coefs)

  x <- data[, data_observed$exposure]
  ystar <- data[, data_observed$outcome]

  force_binary(ystar, "Outcome must be a binary integer.")
  force_len(
    len_u1y0_coefs,
    3 + len_c,
    paste0(
      "Incorrect length of U1Y0 model coefficients. ",
      "Length should equal 3 + number of confounders."
    )
  )
  force_len(
    len_u0y1_coefs,
    3 + len_c,
    paste0(
      "Incorrect length of U0Y1 model coefficients. ",
      "Length should equal 3 + number of confounders."
    )
  )
  force_len(
    len_u1y1_coefs,
    3 + len_c,
    paste0(
      "Incorrect length of U1Y1 model coefficients. ",
      "Length should equal 3 + number of confounders."
    )
  )

  # Extract model coefficients
  u1y0_0 <- u1y0_model_coefs[1]
  u1y0_x <- u1y0_model_coefs[2]
  u1y0_ystar <- u1y0_model_coefs[3]
  u1y0_c <- u1y0_model_coefs[4:len_u1y0_coefs]

  u0y1_0 <- u0y1_model_coefs[1]
  u0y1_x <- u0y1_model_coefs[2]
  u0y1_ystar <- u0y1_model_coefs[3]
  u0y1_c <- u0y1_model_coefs[4:len_u0y1_coefs]

  u1y1_0 <- u1y1_model_coefs[1]
  u1y1_x <- u1y1_model_coefs[2]
  u1y1_ystar <- u1y1_model_coefs[3]
  u1y1_c <- u1y1_model_coefs[4:len_u1y1_coefs]

  # Create base dataframe
  df <- data.frame(X = x, Ystar = ystar)

  # Add confounders if they exist
  if (!is.null(confounders)) {
    for (i in seq_along(confounders)) {
      df[[paste0("C", i)]] <- data[, confounders[i]]
    }
  }

  # Calculate predictions
  u1y0_lp <- u1y0_0 + u1y0_x * df$X + u1y0_ystar * df$Ystar
  u0y1_lp <- u0y1_0 + u0y1_x * df$X + u0y1_ystar * df$Ystar
  u1y1_lp <- u1y1_0 + u1y1_x * df$X + u1y1_ystar * df$Ystar

  if (!is.null(confounders)) {
    C_matrix <- as.matrix(
      df[, paste0("C", seq_along(confounders)), drop = FALSE]
    )
    u1y0_lp <- u1y0_lp + C_matrix %*% u1y0_c
    u0y1_lp <- u0y1_lp + C_matrix %*% u0y1_c
    u1y1_lp <- u1y1_lp + C_matrix %*% u1y1_c
  }

  p_u1y0 <- exp(u1y0_lp)
  p_u0y1 <- exp(u0y1_lp)
  p_u1y1 <- exp(u1y1_lp)

  denom <- (1 + p_u1y0 + p_u0y1 + p_u1y1)

  u0y0_pred <- 1 / denom
  u1y0_pred <- p_u1y0 / denom
  u0y1_pred <- p_u0y1 / denom
  u1y1_pred <- p_u1y1 / denom

  # Create prediction dataframe
  df_uy_pred <- data.frame(
    U0Y0 = u0y0_pred,
    U1Y0 = u1y0_pred,
    U0Y1 = u0y1_pred,
    U1Y1 = u1y1_pred
  )
  df_uy_pred4 <- bind_rows(df_uy_pred, df_uy_pred, df_uy_pred, df_uy_pred)

  # Create combined dataframe with all scenarios
  combined <- bind_rows(df, df, df, df) %>%
    bind_cols(df_uy_pred4) %>%
    mutate(
      Ubar = rep(c(1, 0, 1, 0), each = n),
      Ybar = rep(c(1, 1, 0, 0), each = n),
      pUY = case_when(
        Ubar == 0 & Ybar == 0 ~ U0Y0,
        Ubar == 1 & Ybar == 0 ~ U1Y0,
        Ubar == 0 & Ybar == 1 ~ U0Y1,
        Ubar == 1 & Ybar == 1 ~ U1Y1
      )
    )

  # Construct final model formula
  model_terms <- c("X", "Ubar")
  if (!is.null(confounders)) {
    model_terms <- c(model_terms, paste0("C", seq_along(confounders)))
  }
  model_formula <- as.formula(
    paste("Ybar ~", paste(model_terms, collapse = " + "))
  )

  # Fit final model with weights
  suppressWarnings({
    final <- glm(
      model_formula,
      family = binomial(link = "logit"),
      weights = combined$pUY,
      data = combined
    )
  })

  return(final)
}


adjust_uc_om <- function(
    data_observed,
    data_validation = NULL,
    bias_params = NULL,
    level = 0.95) {
  if (!is.null(data_validation)) {
    final <- adjust_uc_om_val(
      data_observed,
      data_validation
    )
  } else if (!is.null(bias_params)) {
    if (all(c("u", "y") %in% names(bias_params$coef_list))) {
      final <- adjust_uc_om_coef_single(
        data_observed,
        u_model_coefs = bias_params$coef_list$u,
        y_model_coefs = bias_params$coef_list$y
      )
    } else if (
      all(
        c("u1y0", "u0y1", "u1y1") %in%
          names(bias_params$coef_list)
      )
    ) {
      final <- adjust_uc_om_coef_multinom(
        data_observed,
        u1y0_model_coefs = bias_params$coef_list$u1y0,
        u0y1_model_coefs = bias_params$coef_list$u0y1,
        u1y1_model_coefs = bias_params$coef_list$u1y1
      )
    } else {
      (
        stop(
          paste0(
            "bias_params must specify parameters for ",
            "uncontrolled confounding and outcome misclassification"
          ),
          call. = FALSE
        )
      )
    }
  }

  calculate_results(final, level, y_binary = TRUE)
}

#' Simulated data with uncontrolled confounding and outcome misclassification
#'
#' Data containing two sources of bias, three known confounders, and
#' 100,000 observations. This data is obtained from `df_uc_om_source`
#' by removing the columns *Y* and *U*. The resulting data
#' corresponds to what a researcher would see in the real-world: a
#' misclassified outcome, *Ystar*, and missing data on the binary
#' confounder *U*. As seen in `df_uc_omc_source`, the true, unbiased
#' exposure-outcome odds ratio = 2.
#'
#' @format A dataframe with 100,000 rows and 5 columns:
#' \describe{
#'     \item{X}{exposure, 1 = present and 0 = absent}
#'     \item{Ystar}{misclassified outcome, 1 = present and 0 = absent}
#'     \item{C1}{1st confounder, 1 = present and 0 = absent}
#'     \item{C2}{2nd confounder, 1 = present and 0 = absent}
#'     \item{C3}{3rd confounder, 1 = present and 0 = absent}
#' }
"df_uc_om"

#' Data source for `df_uc_om`
#'
#' Data with complete information on the two sources of bias, three known
#' confounders, and 100,000 observations. This data is used to derive
#' `df_uc_om` and can be used to obtain bias parameters for purposes
#' of validating the simultaneous multi-bias adjustment method with
#' `df_uc_om`. With this source data, the fitted regression
#' \ifelse{html}{\out{logit(P(Y=1)) = &alpha;<sub>0</sub> + &alpha;<sub>1</sub>X + &alpha;<sub>2</sub>C1 + &alpha;<sub>3</sub>U}}{\eqn{logit(P(Y=1)) = \alpha_0 + \alpha_1 X + \alpha_2 C1 + \alpha_3 U}}
#' shows that the true, unbiased exposure-outcome odds ratio = 2.
#'
#' @format A dataframe with 100,000 rows and 7 columns:
#' \describe{
#'     \item{X}{exposure, 1 = present and 0 = absent}
#'     \item{Y}{outcome, 1 = present and 0 = absent}
#'     \item{C1}{1st confounder, 1 = present and 0 = absent}
#'     \item{C2}{2nd confounder, 1 = present and 0 = absent}
#'     \item{C3}{3rd confounder, 1 = present and 0 = absent}
#'     \item{U}{unmeasured confounder, 1 = present and 0 = absent}
#'     \item{Ystar}{misclassified outcome, 1 = present and 0 = absent}
#' }
"df_uc_om_source"

# Adjust for uncontrolled confounding, exposure misclassification,
# and selection bias

# the following functions feed into adjust_uc_em_sel():
# adjust_uc_em_sel_val() (data_validation input, method: imputation & weighting),
# adjust_uc_em_sel_coef_single() (bias_params input, method: imputation & weighting),
# adjust_uc_em_sel_coef_multinom() (bias_params input, method: weighting)

adjust_uc_em_sel_val <- function(
    data_observed,
    data_validation) {
  if (!all(data_observed$confounders %in% data_validation$confounders)) {
    stop(
      "All confounders in observed data must be present in validation data.",
      call. = FALSE
    )
  }

  if (
    length(data_validation$confounders) - length(data_observed$confounders) != 1 ||
      is.null(data_validation$misclassified_exposure) ||
      is.null(data_validation$selection)
  ) {
    stop(
      paste0(
        "This function is adjusting for three biases: uncontrolled confounding, exposure misclassification, and selection bias.",
        "\n",
        "Validation data must have: 1) one more confounder than the observed data, 2) a true and misclassified exposure specified, 3) a selection indicator column specified."
      ),
      call. = FALSE
    )
  }

  n <- nrow(data_observed$data)

  df <- data.frame(
    Xstar = data_observed$data[, data_observed$exposure],
    Y = data_observed$data[, data_observed$outcome]
  )
  df <- bind_cols(
    df,
    data_observed$data %>%
      select(all_of(data_observed$confounders))
  )

  if (all(df$Y %in% 0:1)) {
    y_binary <- TRUE
  } else {
    y_binary <- FALSE
  }

  df_val <- data.frame(
    X = data_validation$data[, data_validation$true_exposure],
    Y = data_validation$data[, data_validation$true_outcome],
    Xstar = data_validation$data[, data_validation$misclassified_exposure],
    S = data_validation$data[, data_validation$selection]
  )

  uc <- setdiff(data_validation$confounders, data_observed$confounders)
  df_val$U <- data_validation$data[, uc]
  df_val <- bind_cols(
    df_val,
    data_validation$data %>%
      select(all_of(data_observed$confounders))
  )

  force_match(
    df$Y,
    df_val$Y,
    "Outcomes from both datasets must both be binary or both be continuous."
  )
  force_binary(
    df_val$U,
    "Uncontrolled confounder in validation data must be a binary integer."
  )
  force_binary(
    df$Xstar,
    "Exposure in observed data must be a binary integer."
  )
  force_binary(
    df_val$Xstar,
    "Misclassified exposure in validation data must be a binary integer."
  )
  force_binary(
    df_val$X,
    "True exposure in validation data must be a binary integer."
  )
  force_binary(
    df_val$S,
    "Selection indicator in validation data must be a binary integer."
  )

  x_mod <- glm(X ~ Xstar + Y + . - U - S,
    family = binomial(link = "logit"),
    data = df_val
  )

  x_mod_coefs <- coef(x_mod)
  x_mod_se <- sqrt(diag(vcov(x_mod)))

  # Sample coefficients independently using their standard errors
  x_mod_coefs_sampled <- rnorm(length(x_mod_coefs), x_mod_coefs, x_mod_se)
  x_pred <- x_mod_coefs_sampled[1]

  for (i in 2:length(x_mod_coefs_sampled)) {
    var_name <- names(x_mod_coefs)[i]
    x_pred <- x_pred + df[[var_name]] * x_mod_coefs_sampled[i]
  }

  df$Xpred <- rbinom(n, 1, plogis(x_pred))

  u_mod <- glm(U ~ X + Y,
    family = binomial(link = "logit"),
    data = df_val
  )

  u_mod_coefs <- coef(u_mod)
  u_mod_se <- sqrt(diag(vcov(u_mod)))

  # Sample coefficients independently using their standard errors
  u_mod_coefs_sampled <- rnorm(length(u_mod_coefs), u_mod_coefs, u_mod_se)
  u_pred <- u_mod_coefs_sampled[1]

  for (i in 2:length(u_mod_coefs_sampled)) {
    var_name <- names(u_mod_coefs)[i]
    var_name <- gsub("X", "Xpred", var_name) # col X is not in df
    u_pred <- u_pred + df[[var_name]] * u_mod_coefs_sampled[i]
  }

  df$Upred <- rbinom(n, 1, plogis(u_pred))

  s_mod <- glm(S ~ Xstar + Y + . - X - U,
    family = binomial(link = "logit"),
    data = df_val
  )

  s_mod_coefs <- coef(s_mod)
  s_mod_se <- sqrt(diag(vcov(s_mod)))

  # Sample coefficients independently using their standard errors
  s_mod_coefs_sampled <- rnorm(length(s_mod_coefs), s_mod_coefs, s_mod_se)
  s_pred <- s_mod_coefs_sampled[1]

  for (i in 2:length(s_mod_coefs_sampled)) {
    var_name <- names(s_mod_coefs)[i]
    s_pred <- s_pred + df[[var_name]] * s_mod_coefs_sampled[i]
  }

  df$Spred <- plogis(s_pred)

  if (y_binary) {
    suppressWarnings({
      final <- glm(
        Y ~ Xpred + Upred + . - Xstar - Spred,
        family = binomial(link = "logit"),
        weights = (1 / df$Spred),
        data = df
      )
    })
  } else {
    suppressWarnings({
      final <- lm(
        Y ~ Xpred + Upred + . - Xstar - Spred,
        weights = (1 / df$Spred),
        data = df
      )
    })
  }

  return(final)
}


adjust_uc_em_sel_coef_single <- function(
    data_observed,
    u_model_coefs,
    x_model_coefs,
    s_model_coefs) {
  data <- data_observed$data
  n <- nrow(data)
  confounders <- data_observed$confounders
  len_c <- length(confounders)
  len_u_coefs <- length(u_model_coefs)
  len_x_coefs <- length(x_model_coefs)
  len_s_coefs <- length(s_model_coefs)

  xstar <- data[, data_observed$exposure]
  y <- data[, data_observed$outcome]

  force_binary(xstar, "Exposure must be a binary integer.")
  force_len(
    len_u_coefs,
    3,
    paste0(
      "Incorrect length of U model coefficients. ",
      "Length should equal 3."
    )
  )
  force_len(
    len_x_coefs,
    3 + len_c,
    paste0(
      "Incorrect length of X model coefficients. ",
      "Length should equal 3 + number of confounders."
    )
  )
  force_len(
    len_s_coefs,
    3 + len_c,
    paste0(
      "Incorrect length of S model coefficients. ",
      "Length should equal 3 + number of confounders."
    )
  )

  if (all(y %in% 0:1)) {
    y_binary <- TRUE
  } else {
    y_binary <- FALSE
  }

  # Extract model coefficients
  u1_0 <- u_model_coefs[1]
  u1_x <- u_model_coefs[2]
  u1_y <- u_model_coefs[3]

  x1_0 <- x_model_coefs[1]
  x1_xstar <- x_model_coefs[2]
  x1_y <- x_model_coefs[3]
  x1_c <- x_model_coefs[4:len_x_coefs]

  s1_0 <- s_model_coefs[1]
  s1_xstar <- s_model_coefs[2]
  s1_y <- s_model_coefs[3]
  s1_c <- s_model_coefs[4:len_s_coefs]

  # Create base dataframe
  df <- data.frame(Xstar = xstar, Y = y)

  # Add confounders if they exist
  if (!is.null(confounders)) {
    for (i in seq_along(confounders)) {
      df[[paste0("C", i)]] <- data[, confounders[i]]
    }
  }

  # Calculate X predictions
  x_lp <- x1_0 + x1_xstar * df$Xstar + x1_y * df$Y
  if (!is.null(confounders)) {
    C_matrix <- as.matrix(
      df[, paste0("C", seq_along(confounders)), drop = FALSE]
    )
    x_lp <- x_lp + C_matrix %*% x1_c
  }
  df$Xpred <- rbinom(n, 1, plogis(x_lp))

  # Calculate U predictions
  df$Upred <- rbinom(n, 1, plogis(u1_0 + u1_x * df$Xpred + u1_y * df$Y))

  # Calculate S predictions
  s_lp <- s1_0 + s1_xstar * df$Xstar + s1_y * df$Y
  if (!is.null(confounders)) {
    C_matrix <- as.matrix(
      df[, paste0("C", seq_along(confounders)), drop = FALSE]
    )
    s_lp <- s_lp + C_matrix %*% s1_c
  }
  df$pS <- plogis(s_lp)

  # Construct final model formula
  model_terms <- c("Xpred", "Upred")
  if (!is.null(confounders)) {
    model_terms <- c(model_terms, paste0("C", seq_along(confounders)))
  }
  model_formula <- as.formula(
    paste("Y ~", paste(model_terms, collapse = " + "))
  )

  # Fit final model
  if (y_binary) {
    suppressWarnings({
      final <- glm(
        model_formula,
        family = binomial(link = "logit"),
        weights = (1 / df$pS),
        data = df
      )
    })
  } else {
    final <- lm(
      model_formula,
      weights = (1 / df$pS),
      data = df
    )
  }

  return(final)
}


adjust_uc_em_sel_coef_multinom <- function(
    data_observed,
    x1u0_model_coefs,
    x0u1_model_coefs,
    x1u1_model_coefs,
    s_model_coefs) {
  data <- data_observed$data
  n <- nrow(data)
  confounders <- data_observed$confounders
  len_c <- length(confounders)
  len_x1u0_coefs <- length(x1u0_model_coefs)
  len_x0u1_coefs <- length(x0u1_model_coefs)
  len_x1u1_coefs <- length(x1u1_model_coefs)
  len_s_coefs <- length(s_model_coefs)

  xstar <- data[, data_observed$exposure]
  y <- data[, data_observed$outcome]

  force_binary(xstar, "Exposure must be a binary integer.")
  force_len(
    len_x1u0_coefs,
    3 + len_c,
    paste0(
      "Incorrect length of X1U0 model coefficients. ",
      "Length should equal 3 + number of confounders."
    )
  )
  force_len(
    len_x0u1_coefs,
    3 + len_c,
    paste0(
      "Incorrect length of X0U1 model coefficients. ",
      "Length should equal 3 + number of confounders."
    )
  )
  force_len(
    len_x1u1_coefs,
    3 + len_c,
    paste0(
      "Incorrect length of X1U1 model coefficients. ",
      "Length should equal 3 + number of confounders."
    )
  )
  force_len(
    len_s_coefs,
    3 + len_c,
    paste0(
      "Incorrect length of S model coefficients. ",
      "Length should equal 3 + number of confounders."
    )
  )

  if (all(y %in% 0:1)) {
    y_binary <- TRUE
  } else {
    y_binary <- FALSE
  }

  # Extract model coefficients
  s1_0 <- s_model_coefs[1]
  s1_xstar <- s_model_coefs[2]
  s1_y <- s_model_coefs[3]
  s1_c <- s_model_coefs[4:len_s_coefs]

  x1u0_0 <- x1u0_model_coefs[1]
  x1u0_xstar <- x1u0_model_coefs[2]
  x1u0_y <- x1u0_model_coefs[3]
  x1u0_c <- x1u0_model_coefs[4:len_x1u0_coefs]

  x0u1_0 <- x0u1_model_coefs[1]
  x0u1_xstar <- x0u1_model_coefs[2]
  x0u1_y <- x0u1_model_coefs[3]
  x0u1_c <- x0u1_model_coefs[4:len_x0u1_coefs]

  x1u1_0 <- x1u1_model_coefs[1]
  x1u1_xstar <- x1u1_model_coefs[2]
  x1u1_y <- x1u1_model_coefs[3]
  x1u1_c <- x1u1_model_coefs[4:len_x1u1_coefs]

  # Create base dataframe
  df <- data.frame(Xstar = xstar, Y = y)

  # Add confounders if they exist
  if (!is.null(confounders)) {
    for (i in seq_along(confounders)) {
      df[[paste0("C", i)]] <- data[, confounders[i]]
    }
  }

  # Calculate predictions
  x1u0_lp <- x1u0_0 + x1u0_xstar * df$Xstar + x1u0_y * df$Y
  x0u1_lp <- x0u1_0 + x0u1_xstar * df$Xstar + x0u1_y * df$Y
  x1u1_lp <- x1u1_0 + x1u1_xstar * df$Xstar + x1u1_y * df$Y

  if (!is.null(confounders)) {
    C_matrix <- as.matrix(
      df[, paste0("C", seq_along(confounders)), drop = FALSE]
    )
    x1u0_lp <- x1u0_lp + C_matrix %*% x1u0_c
    x0u1_lp <- x0u1_lp + C_matrix %*% x0u1_c
    x1u1_lp <- x1u1_lp + C_matrix %*% x1u1_c
  }

  p_x1u0 <- exp(x1u0_lp)
  p_x0u1 <- exp(x0u1_lp)
  p_x1u1 <- exp(x1u1_lp)

  denom <- (1 + p_x1u0 + p_x0u1 + p_x1u1)

  x0u0_pred <- 1 / denom
  x1u0_pred <- p_x1u0 / denom
  x0u1_pred <- p_x0u1 / denom
  x1u1_pred <- p_x1u1 / denom

  # Create prediction dataframe
  df_xu_pred <- data.frame(
    X0U0 = x0u0_pred,
    X1U0 = x1u0_pred,
    X0U1 = x0u1_pred,
    X1U1 = x1u1_pred
  )
  df_xu_pred4 <- bind_rows(df_xu_pred, df_xu_pred, df_xu_pred, df_xu_pred)

  # Create combined dataframe
  combined <- bind_rows(df, df, df, df) %>%
    bind_cols(df_xu_pred4) %>%
    mutate(
      Xbar = rep(c(1, 0, 1, 0), each = n),
      Ubar = rep(c(1, 1, 0, 0), each = n),
      pXU = case_when(
        Xbar == 0 & Ubar == 0 ~ X0U0,
        Xbar == 1 & Ubar == 0 ~ X1U0,
        Xbar == 0 & Ubar == 1 ~ X0U1,
        Xbar == 1 & Ubar == 1 ~ X1U1
      )
    )

  # Calculate S predictions
  s_lp <- s1_0 + s1_xstar * combined$Xstar + s1_y * combined$Y
  if (!is.null(confounders)) {
    C_matrix_combined <- as.matrix(
      combined[, paste0("C", seq_along(confounders)), drop = FALSE]
    )
    s_lp <- s_lp + C_matrix_combined %*% s1_c
  }
  combined$pS <- plogis(s_lp)

  # Construct final model formula
  model_terms <- c("Xbar", "Ubar")
  if (!is.null(confounders)) {
    model_terms <- c(model_terms, paste0("C", seq_along(confounders)))
  }
  model_formula <- as.formula(
    paste("Y ~", paste(model_terms, collapse = " + "))
  )

  # Fit final model
  if (y_binary) {
    suppressWarnings({
      final <- glm(
        model_formula,
        family = binomial(link = "logit"),
        weights = (combined$pXU / combined$pS),
        data = combined
      )
    })
  } else {
    final <- lm(
      model_formula,
      weights = (combined$pXU / combined$pS),
      data = combined
    )
  }

  return(final)
}


adjust_uc_em_sel <- function(
    data_observed,
    data_validation = NULL,
    bias_params = NULL,
    level = 0.95) {
  data <- data_observed$data
  xstar <- data[, data_observed$exposure]
  y <- data[, data_observed$outcome]

  if (all(y %in% 0:1)) {
    y_binary <- TRUE
  } else {
    y_binary <- FALSE
  }

  if (!is.null(data_validation)) {
    final <- adjust_uc_em_sel_val(
      data_observed,
      data_validation
    )
  } else if (!is.null(bias_params)) {
    if (all(c("u", "x", "s") %in% names(bias_params$coef_list))) {
      final <- adjust_uc_em_sel_coef_single(
        data_observed,
        u_model_coefs = bias_params$coef_list$u,
        x_model_coefs = bias_params$coef_list$x,
        s_model_coefs = bias_params$coef_list$s
      )
    } else if (
      all(
        c("x1u0", "x0u1", "x1u1", "s") %in%
          names(bias_params$coef_list)
      )
    ) {
      final <- adjust_uc_em_sel_coef_multinom(
        data_observed,
        x1u0_model_coefs = bias_params$coef_list$x1u0,
        x0u1_model_coefs = bias_params$coef_list$x0u1,
        x1u1_model_coefs = bias_params$coef_list$x1u1,
        s_model_coefs = bias_params$coef_list$s
      )
    } else {
      (
        stop(
          paste0(
            "bias_params must specify parameters for uncontrolled ",
            "confounding, exposure misclassification, and selection bias"
          ),
          call. = FALSE
        )
      )
    }
  }

  calculate_results(final, level, y_binary)
}

# Adjust for exposure misclassification and outcome misclassification

# the following functions feed into adjust_em_om():
# adjust_em_om_val() (data_validation input, method: imputation),
# adjust_em_om_coef_single() (bias_params input, method: imputation),
# adjust_em_om_coef_multinom() (bias_params input, method: weighting)


adjust_em_om_val <- function(
    data_observed,
    data_validation) {
  if (!all(data_observed$confounders %in% data_validation$confounders)) {
    stop("All confounders in observed data must be present in validation data.", call. = FALSE)
  }

  if (
    is.null(data_validation$misclassified_exposure) ||
      is.null(data_validation$misclassified_outcome)
  ) {
    stop(
      paste0(
        "Attempting to adjust for a misclassified exposure and misclassified outcome.",
        "\n",
        "Validation data must include a true exposure, misclassified exposure, true outcome, and misclassified outcome."
      ),
      call. = FALSE
    )
  }

  n <- nrow(data_observed$data)

  df <- data.frame(
    Xstar = data_observed$data[, data_observed$exposure],
    Ystar = data_observed$data[, data_observed$outcome]
  )
  df <- bind_cols(
    df,
    data_observed$data %>%
      select(all_of(data_observed$confounders))
  )

  df_val <- data.frame(
    X = data_validation$data[, data_validation$true_exposure],
    Y = data_validation$data[, data_validation$true_outcome],
    Xstar = data_validation$data[, data_validation$misclassified_exposure],
    Ystar = data_validation$data[, data_validation$misclassified_outcome]
  )
  df_val <- bind_cols(
    df_val,
    data_validation$data %>%
      select(all_of(data_validation$confounders))
  )

  force_binary(
    df$Xstar,
    "Exposure in observed data must be a binary integer."
  )
  force_binary(
    df$Ystar,
    "Outcome in observed data must be a binary integer."
  )
  force_binary(
    df_val$Xstar,
    "Misclassified exposure in validation data must be a binary integer."
  )
  force_binary(
    df_val$X,
    "True exposure in validation data must be a binary integer."
  )
  force_binary(
    df_val$Ystar,
    "Misclassified outcome in validation data must be a binary integer."
  )
  force_binary(
    df_val$Y,
    "True outcome in validation data must be a binary integer."
  )

  x_mod <- glm(X ~ Xstar + Ystar + . - Y,
    family = binomial(link = "logit"),
    data = df_val
  )

  x_mod_coefs <- coef(x_mod)
  x_mod_se <- sqrt(diag(vcov(x_mod)))

  # Sample coefficients independently using their standard errors
  x_mod_coefs_sampled <- rnorm(length(x_mod_coefs), x_mod_coefs, x_mod_se)
  x_pred <- x_mod_coefs_sampled[1]

  for (i in 2:length(x_mod_coefs_sampled)) {
    var_name <- names(x_mod_coefs)[i]
    x_pred <- x_pred + df[[var_name]] * x_mod_coefs_sampled[i]
  }

  df$Xpred <- rbinom(n, 1, plogis(x_pred))

  y_mod <- glm(Y ~ X + Ystar + . - Xstar,
    family = binomial(link = "logit"),
    data = df_val
  )

  y_mod_coefs <- coef(y_mod)
  y_mod_se <- sqrt(diag(vcov(y_mod)))

  # Sample coefficients independently using their standard errors
  y_mod_coefs_sampled <- rnorm(length(y_mod_coefs), y_mod_coefs, y_mod_se)
  y_pred <- y_mod_coefs_sampled[1]

  for (i in 2:length(y_mod_coefs_sampled)) {
    var_name <- names(y_mod_coefs)[i]
    var_name <- gsub("X", "Xpred", var_name) # col X is not in df
    y_pred <- y_pred + df[[var_name]] * y_mod_coefs_sampled[i]
  }

  df$Ypred <- rbinom(n, 1, plogis(y_pred))

  final <- glm(
    Ypred ~ Xpred + . - Xstar - Ystar,
    family = binomial(link = "logit"),
    data = df
  )

  return(final)
}


adjust_em_om_coef_single <- function(
    data_observed,
    x_model_coefs,
    y_model_coefs) {
  data <- data_observed$data
  n <- nrow(data)
  confounders <- data_observed$confounders
  len_c <- length(confounders)
  len_x_coefs <- length(x_model_coefs)
  len_y_coefs <- length(y_model_coefs)

  xstar <- data[, data_observed$exposure]
  ystar <- data[, data_observed$outcome]

  force_binary(xstar, "Exposure must be a binary integer.")
  force_binary(ystar, "Outcome must be a binary integer.")
  force_len(
    len_x_coefs,
    3 + len_c,
    paste0(
      "Incorrect length of X model coefficients. ",
      "Length should equal 3 + number of confounders."
    )
  )
  force_len(
    len_y_coefs,
    3 + len_c,
    paste0(
      "Incorrect length of Y model coefficients. ",
      "Length should equal 3 + number of confounders."
    )
  )

  # Extract model coefficients
  x1_0 <- x_model_coefs[1]
  x1_xstar <- x_model_coefs[2]
  x1_ystar <- x_model_coefs[3]
  x1_c <- x_model_coefs[4:len_x_coefs]

  y1_0 <- y_model_coefs[1]
  y1_x <- y_model_coefs[2]
  y1_ystar <- y_model_coefs[3]
  y1_c <- y_model_coefs[4:len_y_coefs]

  # Create base dataframe
  df <- data.frame(Xstar = xstar, Ystar = ystar)

  # Add confounders if they exist
  if (!is.null(confounders)) {
    for (i in seq_along(confounders)) {
      df[[paste0("C", i)]] <- data[, confounders[i]]
    }
  }

  # Calculate X predictions
  x_lp <- x1_0 + x1_xstar * df$Xstar + x1_ystar * df$Ystar
  if (!is.null(confounders)) {
    C_matrix <- as.matrix(
      df[, paste0("C", seq_along(confounders)), drop = FALSE]
    )
    x_lp <- x_lp + C_matrix %*% x1_c
  }
  df$Xpred <- rbinom(n, 1, plogis(x_lp))

  # Calculate Y predictions
  y_lp <- y1_0 + y1_x * df$Xpred + y1_ystar * df$Ystar
  if (!is.null(confounders)) {
    C_matrix <- as.matrix(
      df[, paste0("C", seq_along(confounders)), drop = FALSE]
    )
    y_lp <- y_lp + C_matrix %*% y1_c
  }
  df$Ypred <- rbinom(n, 1, plogis(y_lp))

  # Construct final model formula
  model_terms <- c("Xpred")
  if (!is.null(confounders)) {
    model_terms <- c(model_terms, paste0("C", seq_along(confounders)))
  }
  model_formula <- as.formula(
    paste("Ypred ~", paste(model_terms, collapse = " + "))
  )

  # Fit final model
  final <- glm(
    model_formula,
    family = binomial(link = "logit"),
    data = df
  )

  return(final)
}


adjust_em_om_coef_multinom <- function(
    data_observed,
    x1y0_model_coefs,
    x0y1_model_coefs,
    x1y1_model_coefs) {
  data <- data_observed$data
  n <- nrow(data)
  confounders <- data_observed$confounders
  len_c <- length(confounders)
  len_x1y0_coefs <- length(x1y0_model_coefs)
  len_x0y1_coefs <- length(x0y1_model_coefs)
  len_x1y1_coefs <- length(x1y1_model_coefs)

  xstar <- data[, data_observed$exposure]
  ystar <- data[, data_observed$outcome]

  force_binary(xstar, "Exposure must be a binary integer.")
  force_binary(ystar, "Outcome must be a binary integer.")
  force_len(
    len_x1y0_coefs,
    3 + len_c,
    paste0(
      "Incorrect length of X1Y0 model coefficients. ",
      "Length should equal 3 + number of confounders."
    )
  )
  force_len(
    len_x0y1_coefs,
    3 + len_c,
    paste0(
      "Incorrect length of X0Y1 model coefficients. ",
      "Length should equal 3 + number of confounders."
    )
  )
  force_len(
    len_x1y1_coefs,
    3 + len_c,
    paste0(
      "Incorrect length of X1Y1 model coefficients. ",
      "Length should equal 3 + number of confounders."
    )
  )

  # Extract model coefficients
  x1y0_0 <- x1y0_model_coefs[1]
  x1y0_xstar <- x1y0_model_coefs[2]
  x1y0_ystar <- x1y0_model_coefs[3]
  x1y0_c <- x1y0_model_coefs[4:len_x1y0_coefs]

  x0y1_0 <- x0y1_model_coefs[1]
  x0y1_xstar <- x0y1_model_coefs[2]
  x0y1_ystar <- x0y1_model_coefs[3]
  x0y1_c <- x0y1_model_coefs[4:len_x0y1_coefs]

  x1y1_0 <- x1y1_model_coefs[1]
  x1y1_xstar <- x1y1_model_coefs[2]
  x1y1_ystar <- x1y1_model_coefs[3]
  x1y1_c <- x1y1_model_coefs[4:len_x1y1_coefs]

  # Create base dataframe
  df <- data.frame(Xstar = xstar, Ystar = ystar)

  # Add confounders if they exist
  if (!is.null(confounders)) {
    for (i in seq_along(confounders)) {
      df[[paste0("C", i)]] <- data[, confounders[i]]
    }
  }

  # Calculate predictions
  x1y0_lp <- x1y0_0 + x1y0_xstar * df$Xstar + x1y0_ystar * df$Ystar
  x0y1_lp <- x0y1_0 + x0y1_xstar * df$Xstar + x0y1_ystar * df$Ystar
  x1y1_lp <- x1y1_0 + x1y1_xstar * df$Xstar + x1y1_ystar * df$Ystar

  if (!is.null(confounders)) {
    C_matrix <- as.matrix(
      df[, paste0("C", seq_along(confounders)), drop = FALSE]
    )
    x1y0_lp <- x1y0_lp + C_matrix %*% x1y0_c
    x0y1_lp <- x0y1_lp + C_matrix %*% x0y1_c
    x1y1_lp <- x1y1_lp + C_matrix %*% x1y1_c
  }

  p_x1y0 <- exp(x1y0_lp)
  p_x0y1 <- exp(x0y1_lp)
  p_x1y1 <- exp(x1y1_lp)

  denom <- (1 + p_x1y0 + p_x0y1 + p_x1y1)

  x0y0_pred <- 1 / denom
  x1y0_pred <- p_x1y0 / denom
  x0y1_pred <- p_x0y1 / denom
  x1y1_pred <- p_x1y1 / denom

  # Create prediction dataframe
  df_xy_pred <- data.frame(
    X0Y0 = x0y0_pred,
    X1Y0 = x1y0_pred,
    X0Y1 = x0y1_pred,
    X1Y1 = x1y1_pred
  )
  df_xy_pred4 <- bind_rows(df_xy_pred, df_xy_pred, df_xy_pred, df_xy_pred)

  # Create combined dataframe with all scenarios
  combined <- bind_rows(df, df, df, df) %>%
    bind_cols(df_xy_pred4) %>%
    mutate(
      Xbar = rep(c(1, 0, 1, 0), each = n),
      Ybar = rep(c(1, 1, 0, 0), each = n),
      pXY = case_when(
        Xbar == 0 & Ybar == 0 ~ X0Y0,
        Xbar == 1 & Ybar == 0 ~ X1Y0,
        Xbar == 0 & Ybar == 1 ~ X0Y1,
        Xbar == 1 & Ybar == 1 ~ X1Y1
      )
    )

  # Construct final model formula
  model_terms <- c("Xbar")
  if (!is.null(confounders)) {
    model_terms <- c(model_terms, paste0("C", seq_along(confounders)))
  }
  model_formula <- as.formula(
    paste("Ybar ~", paste(model_terms, collapse = " + "))
  )

  # Fit final model with weights
  suppressWarnings({
    final <- glm(
      model_formula,
      family = binomial(link = "logit"),
      weights = combined$pXY,
      data = combined
    )
  })

  return(final)
}


adjust_em_om <- function(
    data_observed,
    data_validation = NULL,
    bias_params = NULL,
    level = 0.95) {
  if (!is.null(data_validation)) {
    final <- adjust_em_om_val(
      data_observed,
      data_validation
    )
  } else if (!is.null(bias_params)) {
    if (all(c("x", "y") %in% names(bias_params$coef_list))) {
      final <- adjust_em_om_coef_single(
        data_observed,
        x_model_coefs = bias_params$coef_list$x,
        y_model_coefs = bias_params$coef_list$y
      )
    } else if (
      all(
        c("x1y0", "x0y1", "x1y1") %in%
          names(bias_params$coef_list)
      )
    ) {
      final <- adjust_em_om_coef_multinom(
        data_observed,
        x1y0_model_coefs = bias_params$coef_list$x1y0,
        x0y1_model_coefs = bias_params$coef_list$x0y1,
        x1y1_model_coefs = bias_params$coef_list$x1y1
      )
    } else {
      (
        stop(
          paste0(
            "bias_params must specify parameters for ",
            "exposure misclassification and outcome misclassification"
          ),
          call. = FALSE
        )
      )
    }
  }

  calculate_results(final, level, y_binary = TRUE)
}

#' Simultaneously adjust for multiple biases
#'
#' `multibias_adjust` returns the exposure-outcome odds ratio and confidence
#' interval, adjusted for one or more biases.
#'
#' Bias adjustment can be performed by inputting either a validation dataset or
#' the necessary bias parameters. Values for the bias parameters
#' can be applied as fixed values or as single draws from a probability
#' distribution (ex: `rnorm(1, mean = 2, sd = 1)`). The latter has
#' the advantage of allowing the researcher to capture the uncertainty
#' in the bias parameter estimates. To incorporate this uncertainty in the
#' estimate and confidence interval, this function should be run in loop across
#' bootstrap samples of the dataframe for analysis. The estimate and
#' confidence interval would then be obtained from the median and quantiles
#' of the distribution of odds ratio estimates.
#'
#' @param data_observed Object of class `data_observed` corresponding to the
#' data to perform bias analysis on.
#' @param data_validation Object of class `data_validation` corresponding to
#' the validation data used to adjust for bias in the observed data. The
#' validation data should have data for the same variables as in
#' `data_observed`, plus data for the missing variables leading to bias.
#' @param bias_params Object of class 'bias_params' corresponding to the
#' bias parameters used to adjust for bias in the observed data. There must
#' be parameters corresponding to the bias or biases specified in
#' `data_observed`.
#' @param bootstrap Boolean for whether to perform bootstrapping to obtain
#' the estimate and confidence interval.
#' @param bootstrap_reps Integer number of bootstrap samples to run in
#' bootstrapping.
#' @param level Value from 0-1 representing the full range of the confidence
#' interval. Default is 0.95.
#'
#' @return A list including: the bias-adjusted effect estimate of the exposure
#' on the outcome, the standard error, and the confidence interval as the
#' vector: (lower bound, upper bound).
#'
#' @examples
#' # Adjust for exposure misclassification -------------------------------------
#' df_observed <- data_observed(
#'   data = df_em,
#'   bias = "em",
#'   exposure = "Xstar",
#'   outcome = "Y",
#'   confounders = "C1"
#' )
#'
#' # Using validation data
#' df_validation <- data_validation(
#'   data = df_em_source,
#'   true_exposure = "X",
#'   true_outcome = "Y",
#'   confounders = "C1",
#'   misclassified_exposure = "Xstar"
#' )
#'
#' multibias_adjust(
#'   data_observed = df_observed,
#'   data_validation = df_validation
#' )
#'
#' # Using bias_params
#' bp <- bias_params(coef_list = list(x = c(-2.10, 1.62, 0.63, 0.35)))
#'
#' multibias_adjust(
#'   data_observed = df_observed,
#'   bias_params = bp
#' )
#'
#' # Adjust for three biases ---------------------------------------------------
#' df_observed <- data_observed(
#'   data = df_uc_om_sel,
#'   bias = c("uc", "om", "sel"),
#'   exposure = "X",
#'   outcome = "Ystar",
#'   confounders = c("C1", "C2", "C3")
#' )
#'
#' # Using validation data
#' df_validation <- data_validation(
#'   data = df_uc_om_sel_source,
#'   true_exposure = "X",
#'   true_outcome = "Y",
#'   confounders = c("C1", "C2", "C3", "U"),
#'   misclassified_outcome = "Ystar",
#'   selection = "S"
#' )
#'
#' multibias_adjust(
#'   data_observed = df_observed,
#'   data_validation = df_validation
#' )
#'
#' # Using bias_params
#' bp1 <- bias_params(
#'   coef_list = list(
#'     u = c(-0.32, 0.59, 0.69),
#'     y = c(-2.85, 0.71, 1.63, 0.40, -0.85, 0.22),
#'     s = c(0.00, 0.74, 0.19, 0.02, -0.06, 0.02)
#'   )
#' )
#'
#' multibias_adjust(
#'   data_observed = df_observed,
#'   bias_params = bp1
#' )
#'
#' bp2 <- bias_params(
#'   coef_list = list(
#'     u1y0 = c(-0.20, 0.62, 0.01, -0.08, 0.10, -0.15),
#'     u0y1 = c(-3.28, 0.63, 1.65, 0.42, -0.85, 0.26),
#'     u1y1 = c(-2.70, 1.22, 1.64, 0.32, -0.77, 0.09),
#'     s = c(0.00, 0.74, 0.19, 0.02, -0.06, 0.02)
#'   )
#' )
#'
#' # with bootstrapping
#' \dontrun{
#' multibias_adjust(
#'   data_observed = df_observed,
#'   bias_params = bp2,
#'   bootstrap = TRUE,
#'   bootstrap_reps = 1000
#' )
#' }
#'
#' @import dplyr
#' @importFrom magrittr %>%
#' @importFrom stats as.formula binomial coef lm median glm plogis rbinom
#' @importFrom stats rnorm sd qnorm quantile vcov
#' @importFrom rlang .data
#'
#' @export

multibias_adjust <- function(
    data_observed,
    data_validation = NULL,
    bias_params = NULL,
    bootstrap = FALSE,
    bootstrap_reps = 100,
    level = 0.95) {
  check_inputs2(data_validation, bias_params)

  bias_dict <- list(
    "uc" = adjust_uc,
    "em" = adjust_em,
    "om" = adjust_om,
    "sel" = adjust_sel,
    "em_uc" = adjust_uc_em,
    "om_uc" = adjust_uc_om,
    "sel_uc" = adjust_uc_sel,
    "em_om" = adjust_em_om,
    "em_sel" = adjust_em_sel,
    "om_sel" = adjust_om_sel,
    "em_sel_uc" = adjust_uc_em_sel,
    "om_sel_uc" = adjust_uc_om_sel
  )

  bias_key <- paste(sort(data_observed$bias), collapse = "_")
  adjust_fn <- bias_dict[[bias_key]]

  if (bootstrap == FALSE) {
    output <- adjust_fn(data_observed, data_validation, bias_params, level)
  } else if (bootstrap == TRUE) {
    df <- data_observed$data
    n <- nrow(df)
    est <- numeric(bootstrap_reps)

    indices <- matrix(
      sample.int(n, n * bootstrap_reps, replace = TRUE),
      nrow = n,
      ncol = bootstrap_reps
    )

    for (i in seq_len(bootstrap_reps)) {
      data_observed$data <- df[indices[, i], ]
      final <- adjust_fn(
        data_observed,
        data_validation,
        bias_params
      )
      est[i] <- final$estimate
    }

    alpha <- 1 - level
    output <- list(
      estimate = median(est),
      std.error = sd(est),
      ci = as.vector(quantile(est, c(alpha / 2, 1 - alpha / 2)))
    )
  }

  return(output)
}

# Adjust for uncontrolled confounding

# the following functions feed into adjust_uc():
# adjust_uc_val() (data_validation input, method: imputation),
# adjust_uc_coef() (bias_params input, method: imputation)

adjust_uc_val <- function(
    data_observed,
    data_validation) {
  if (!all(data_observed$confounders %in% data_validation$confounders)) {
    stop(
      "All confounders in observed data must be present in validation data.",
      call. = FALSE
    )
  }

  if (
    length(data_validation$confounders) - length(data_observed$confounders) != 1
  ) {
    stop(
      paste0(
        "Attempting to adjust for unobserved confounding from one confounder.",
        "\n",
        "Validation data must have one more confounder than the observed data."
      ),
      call. = FALSE
    )
  }

  n <- nrow(data_observed$data)

  df <- data.frame(
    X = data_observed$data[, data_observed$exposure],
    Y = data_observed$data[, data_observed$outcome]
  )
  df <- bind_cols(
    df,
    data_observed$data %>%
      select(all_of(data_observed$confounders))
  )

  if (all(df$Y %in% 0:1)) {
    y_binary <- TRUE
  } else {
    y_binary <- FALSE
  }

  df_val <- data.frame(
    X = data_validation$data[, data_validation$true_exposure],
    Y = data_validation$data[, data_validation$true_outcome]
  )

  uc <- setdiff(data_validation$confounders, data_observed$confounders)
  df_val$U <- data_validation$data[, uc]
  df_val <- bind_cols(
    df_val,
    data_validation$data %>%
      select(all_of(data_observed$confounders))
  )

  force_match(
    df$X,
    df_val$X,
    "Exposures from both datasets must both be binary or both be continuous."
  )
  force_match(
    df$Y,
    df_val$Y,
    "Outcomes from both datasets must both be binary or both be continuous."
  )
  force_binary(
    df_val$U,
    "Uncontrolled confounder in validation data must be a binary integer."
  )

  u_mod <- glm(U ~ X + Y + .,
    family = binomial(link = "logit"),
    data = df_val
  )

  u_mod_coefs <- coef(u_mod)
  u_mod_se <- sqrt(diag(vcov(u_mod)))

  # Sample coefficients independently using their standard errors
  u_mod_coefs_sampled <- rnorm(length(u_mod_coefs), u_mod_coefs, u_mod_se)
  u_pred <- u_mod_coefs_sampled[1]

  for (i in 2:length(u_mod_coefs_sampled)) {
    var_name <- names(u_mod_coefs)[i]
    u_pred <- u_pred + df[[var_name]] * u_mod_coefs_sampled[i]
  }

  df$Upred <- rbinom(n, 1, plogis(u_pred))

  if (y_binary) {
    final <- glm(
      Y ~ X + Upred + .,
      family = binomial(link = "logit"),
      data = df
    )
  } else {
    final <- lm(
      Y ~ X + Upred + .,
      data = df
    )
  }

  return(final)
}


adjust_uc_coef <- function(
    data_observed,
    u_model_coefs) {
  data <- data_observed$data
  n <- nrow(data)
  confounders <- data_observed$confounders
  len_c <- length(confounders)
  len_u_coefs <- length(u_model_coefs)

  x <- data[, data_observed$exposure]
  y <- data[, data_observed$outcome]

  force_len(
    len_u_coefs,
    3 + len_c,
    paste0(
      "Incorrect length of U model coefficients. ",
      "Length should equal 3 + number of confounders."
    )
  )

  if (all(y %in% 0:1)) {
    y_binary <- TRUE
  } else {
    y_binary <- FALSE
  }

  # Extract U model coefficients
  u1_0 <- u_model_coefs[1]
  u1_x <- u_model_coefs[2]
  u1_y <- u_model_coefs[3]
  u1_c <- u_model_coefs[4:len_u_coefs]

  # Create base dataframe
  df <- data.frame(X = x, Y = y)

  # Add confounders if they exist
  if (!is.null(confounders)) {
    for (i in seq_along(confounders)) {
      df[[paste0("C", i)]] <- data[, confounders[i]]
    }
  }

  # Calculate U predictions
  u_lp <- u1_0 + u1_x * df$X + u1_y * df$Y
  if (!is.null(confounders)) {
    C_matrix <- as.matrix(
      df[, paste0("C", seq_along(confounders)), drop = FALSE]
    )
    u_lp <- u_lp + C_matrix %*% u1_c
  }
  df$Upred <- rbinom(n, 1, plogis(u_lp))

  # Construct final model formula
  model_terms <- c("X", "Upred")
  if (!is.null(confounders)) {
    model_terms <- c(model_terms, paste0("C", seq_along(confounders)))
  }
  model_formula <- as.formula(
    paste("Y ~", paste(model_terms, collapse = " + "))
  )

  # Fit final model
  if (y_binary) {
    final <- glm(
      model_formula,
      family = binomial(link = "logit"),
      data = df
    )
  } else {
    final <- lm(
      model_formula,
      data = df
    )
  }

  return(final)
}


adjust_uc <- function(
    data_observed,
    data_validation = NULL,
    bias_params = NULL,
    level = 0.95) {
  data <- data_observed$data
  x <- data[, data_observed$exposure]
  y <- data[, data_observed$outcome]

  if (all(y %in% 0:1)) {
    y_binary <- TRUE
  } else {
    y_binary <- FALSE
  }

  if (!is.null(data_validation)) {
    final <- adjust_uc_val(
      data_observed,
      data_validation
    )
  } else if (!is.null(bias_params)) {
    if (is.null(bias_params$coef_list$u)) {
      stop(
        "bias_params must specify parameters for u to adjust for uncontrolled confounding",
        call. = FALSE
      )
    }
    final <- adjust_uc_coef(
      data_observed,
      u_model_coefs = bias_params$coef_list$u
    )
  }

  calculate_results(final, level, y_binary)
}

is_binary <- function(x) {
  if (is.logical(x)) {
    return(TRUE)
  }
  if (!is.numeric(x)) {
    return(FALSE)
  }
  unique_vals <- unique(x)
  length(unique_vals) == 2 &&
    all(near(unique_vals, 0) | near(unique_vals, 1))
}

is_continuous <- function(x) {
  is.numeric(x) && length(unique(x)) > 2
}

force_binary <- function(col, message) {
  if (!all(is_binary(col))) {
    stop(message, call. = FALSE)
  }
}

force_len <- function(len_observed, len_required, message) {
  if (len_observed != len_required) {
    stop(message, call. = FALSE)
  }
}

force_match <- function(col1, col2, message) {
  col1_type <- if_else(
    is_binary(col1),
    "binary",
    if_else(
      is_continuous(col1),
      "continuous",
      "other"
    )
  )
  col2_type <- if_else(
    is_binary(col2),
    "binary",
    if_else(
      is_continuous(col2),
      "continuous",
      "other"
    )
  )
  if (col1_type != col2_type) {
    stop(message, call. = FALSE)
  }
}

is_any_null <- function(x) {
  if (is.list(x) && inherits(x, "data_validation") == FALSE) {
    any(sapply(x, is.null))
  } else {
    is.null(x)
  }
}

is_all_null <- function(x) {
  if (is.list(x) && inherits(x, "data_validation") == FALSE) {
    all(sapply(x, is.null))
  } else {
    is.null(x)
  }
}

check_inputs2 <- function(input1, input2) {
  input1_name <- deparse(substitute(input1))
  input2_name <- deparse(substitute(input2))

  if (!is_any_null(input1)) {
    if (!is_all_null(input2)) {
      stop(
        paste0(
          "No other bias-adjusting parameters should be specified when ",
          input1_name,
          " is used."
        ),
        call. = FALSE
      )
    }
  } else if (!is_any_null(input2)) {
    if (!is_all_null(input1)) {
      stop(
        paste0(
          "No other bias-adjusting parameters should be specified when ",
          input2_name,
          " is used."
        ),
        call. = FALSE
      )
    }
  } else if (all(is_any_null(input1), is_any_null(input2))) {
    stop(
      paste0(
        "One of:\n",
        "1. ", input1_name, "\n",
        "2. ", input2_name, "\n",
        "must be non-null."
      ),
      call. = FALSE
    )
  }
}

check_inputs3 <- function(input1, input2, input3, ignore = NULL) {
  input1_name <- deparse(substitute(input1))
  input2_name <- deparse(substitute(input2))
  input3_name <- deparse(substitute(input3))

  remove_ignore <- function(x, ignore) {
    if (is.list(x) && !inherits(x, "data_validation")) {
      x[sapply(x, function(y) !(identical(y, ignore)))]
    } else {
      x
    }
  }

  if (!is.null(ignore)) {
    input1 <- remove_ignore(input1, ignore)
    input2 <- remove_ignore(input2, ignore)
    input3 <- remove_ignore(input3, ignore)
  }

  if (!is_any_null(input1)) {
    if (!all(is_all_null(input2), is_all_null(input3))) {
      stop(
        paste0(
          "No other bias-adjusting parameters should be specified when ",
          input1_name,
          " is used."
        ),
        call. = FALSE
      )
    }
  } else if (!is_any_null(input2)) {
    if (!all(is_all_null(input1), is_all_null(input3))) {
      stop(
        paste0(
          "No other bias-adjusting parameters should be specified when ",
          input2_name,
          " is used."
        ),
        call. = FALSE
      )
    }
  } else if (!is_any_null(input3)) {
    if (!all(is_all_null(input1), is_all_null(input2))) {
      stop(
        paste0(
          "No other bias-adjusting parameters should be specified when ",
          input3_name,
          " is used."
        ),
        call. = FALSE
      )
    }
  } else if (
    all(
      is_any_null(input1),
      is_any_null(input2),
      is_any_null(input3)
    )
  ) {
    stop(
      paste0(
        "One of:\n",
        "1. ", input1_name, "\n",
        "2. ", input2_name, "\n",
        "3. ", input3_name, "\n",
        "must be non-null."
      ),
      call. = FALSE
    )
  }
}

calculate_results <- function(final, level, y_binary) {
  est <- summary(final)$coef[2, 1]
  se <- summary(final)$coef[2, 2]
  alpha <- 1 - level

  if (y_binary) {
    estimate <- exp(est)
    ci <- c(
      exp(est + se * qnorm(alpha / 2)),
      exp(est + se * qnorm(1 - alpha / 2))
    )
  } else {
    estimate <- est
    ci <- c(
      est + se * qnorm(alpha / 2),
      est + se * qnorm(1 - alpha / 2)
    )
  }

  return(list(estimate = estimate, std.error = se, ci = ci))
}

# Adjust for outcome misclassification

# the following functions feed into adjust_om():
# adjust_om_val() (data_validation input),
# adjust_om_coef() (bias_params input)

adjust_om_val <- function(
    data_observed,
    data_validation) {
  if (!all(data_observed$confounders %in% data_validation$confounders)) {
    stop(
      "All confounders in observed data must be present in validation data.",
      call. = FALSE
    )
  }

  if (is.null(data_validation$misclassified_outcome)) {
    stop(
      paste0(
        "Attempting to adjust for a misclassified outcome.",
        "\n",
        "Validation data must have a true and misclassified outcome specified."
      ),
      call. = FALSE
    )
  }

  n <- nrow(data_observed$data)

  df <- data.frame(
    X = data_observed$data[, data_observed$exposure],
    Ystar = data_observed$data[, data_observed$outcome]
  )
  df <- bind_cols(
    df,
    data_observed$data %>%
      select(all_of(data_observed$confounders))
  )

  df_val <- data.frame(
    X = data_validation$data[, data_validation$true_exposure],
    Y = data_validation$data[, data_validation$true_outcome],
    Ystar = data_validation$data[, data_validation$misclassified_outcome]
  )
  df_val <- bind_cols(
    df_val,
    data_validation$data %>%
      select(all_of(data_validation$confounders))
  )

  force_match(
    df$X,
    df_val$X,
    "Exposures from both datasets must both be binary or both be continuous."
  )
  force_binary(
    df$Ystar,
    "Outcome in observed data must be a binary integer."
  )
  force_binary(
    df_val$Ystar,
    "Misclassified outcome in validation data must be a binary integer."
  )
  force_binary(
    df_val$Y,
    "True outcome in validation data must be a binary integer."
  )

  y_mod <- glm(Y ~ X + Ystar + .,
    family = binomial(link = "logit"),
    data = df_val
  )

  y_mod_coefs <- coef(y_mod)
  y_mod_se <- sqrt(diag(vcov(y_mod)))

  # Sample coefficients independently using their standard errors
  y_mod_coefs_sampled <- rnorm(length(y_mod_coefs), y_mod_coefs, y_mod_se)
  y_pred <- y_mod_coefs_sampled[1]

  for (i in 2:length(y_mod_coefs_sampled)) {
    var_name <- names(y_mod_coefs)[i]
    y_pred <- y_pred + df[[var_name]] * y_mod_coefs_sampled[i]
  }

  df$Ypred <- rbinom(n, 1, plogis(y_pred))

  final <- glm(
    Ypred ~ X + . - Ystar,
    family = binomial(link = "logit"),
    data = df
  )

  return(final)
}


adjust_om_coef <- function(
    data_observed,
    y_model_coefs) {
  data <- data_observed$data
  n <- nrow(data)
  confounders <- data_observed$confounders
  len_c <- length(confounders)
  len_y_coefs <- length(y_model_coefs)

  x <- data[, data_observed$exposure]
  ystar <- data[, data_observed$outcome]

  force_binary(ystar, "Outcome must be a binary integer.")
  force_len(
    len_y_coefs,
    3 + len_c,
    paste0(
      "Incorrect length of Y model coefficients. ",
      "Length should equal 3 + number of confounders."
    )
  )

  # Extract Y model coefficients
  y1_0 <- y_model_coefs[1]
  y1_x <- y_model_coefs[2]
  y1_ystar <- y_model_coefs[3]
  y1_c <- y_model_coefs[4:len_y_coefs]

  # Create base dataframe
  df <- data.frame(X = x, Ystar = ystar)

  # Add confounders if they exist
  if (!is.null(confounders)) {
    for (i in seq_along(confounders)) {
      df[[paste0("C", i)]] <- data[, confounders[i]]
    }
  }

  # Calculate Y predictions
  y_lp <- y1_0 + y1_x * df$X + y1_ystar * df$Ystar
  if (!is.null(confounders)) {
    C_matrix <- as.matrix(
      df[, paste0("C", seq_along(confounders)), drop = FALSE]
    )
    y_lp <- y_lp + C_matrix %*% y1_c
  }
  df$Ypred <- rbinom(n, 1, plogis(y_lp))

  # Construct final model formula
  model_terms <- c("X")
  if (!is.null(confounders)) {
    model_terms <- c(model_terms, paste0("C", seq_along(confounders)))
  }
  model_formula <- as.formula(
    paste("Ypred ~", paste(model_terms, collapse = " + "))
  )

  # Fit final model
  final <- glm(
    model_formula,
    family = binomial(link = "logit"),
    data = df
  )

  return(final)
}


adjust_om <- function(
    data_observed,
    data_validation = NULL,
    bias_params = NULL,
    level = 0.95) {
  if (!is.null(data_validation)) {
    final <- adjust_om_val(
      data_observed,
      data_validation
    )
  } else if (!is.null(bias_params)) {
    if (is.null(bias_params$coef_list$y)) {
      stop(
        "bias_params must specify parameters for y to adjust for outcome misclassification",
        call. = FALSE
      )
    }
    final <- adjust_om_coef(
      data_observed,
      y_model_coefs = bias_params$coef_list$y
    )
  }

  calculate_results(final, level, y_binary = TRUE)
}

#' Simulated data with selection bias
#'
#' Data containing one source of bias, three known confounders, and 100,000
#' observations. This data is obtained by sampling with replacement with
#' probability = *S* from `df_sel_source` then removing the *S*
#' column. The resulting data corresponds to what a researcher would see
#' in the real-world: missing data for those not selected into the study
#' (*S*=0). As seen in `df_sel_source`, the true, unbiased
#' exposure-outcome odds ratio = 2.
#'
#' @format A dataframe with 100,000 rows and 5 columns:
#' \describe{
#'     \item{X}{exposure, 1 = present and 0 = absent}
#'     \item{Y}{outcome, 1 = present and 0 = absent}
#'     \item{C1}{1st confounder, 1 = present and 0 = absent}
#'     \item{C2}{2nd confounder, 1 = present and 0 = absent}
#'     \item{C3}{3rd confounder, 1 = present and 0 = absent}
#' }
"df_sel"

#' Data source for `df_sel`
#'
#' Data with complete information on study selection, three known
#' confounders, and 100,000 observations. This data is used to derive
#' `df_sel` and can be used to obtain bias parameters for purposes
#' of validating the simultaneous multi-bias adjustment method with
#' `df_sel`. With this source data, the fitted regression
#' \ifelse{html}{\out{logit(P(Y=1)) = &alpha;<sub>0</sub> + &alpha;<sub>1</sub>X + &alpha;<sub>2</sub>C1 + &alpha;<sub>3</sub>C2 + &alpha;<sub>4</sub>C3}}{\eqn{logit(P(Y=1)) = \alpha_0 + \alpha_1 X + \alpha_2 C1 + \alpha_3 C2 + \alpha_4 C3}}
#' shows that the true, unbiased exposure-outcome odds ratio = 2.
#'
#' @format A dataframe with 100,000 rows and 6 columns:
#' \describe{
#'     \item{X}{true exposure, 1 = present and 0 = absent}
#'     \item{Y}{outcome, 1 = present and 0 = absent}
#'     \item{C1}{1st confounder, 1 = present and 0 = absent}
#'     \item{C2}{2nd confounder, 1 = present and 0 = absent}
#'     \item{C3}{3rd confounder, 1 = present and 0 = absent}
#'     \item{S}{selection, 1 = selected into the study and 0 = not selected into the study}
#' }
"df_sel_source"

#' Create a Forest Plot comparing observed and adjusted effect estimates
#'
#' This function generates a forest plot comparing the observed effect estimate
#' with adjusted effect estimates from sensitivity analyses. The plot includes
#' point estimates and confidence intervals for each analysis.
#'
#' @param data_observed Object of class `data_observed` representing the
#' observed causal data and effect of interest.
#' @param multibias_result_list A named list of sensitivity analysis results.
#' Each element should be a result from [multibias_adjust()].
#' @param log_scale Boolean indicating whether to display the x-axis on the
#' log scale. Default is FALSE.
#'
#' @return A ggplot object showing a forest plot with:
#'   \itemize{
#'     \item Point estimates (blue dots)
#'     \item Confidence intervals (gray horizontal lines)
#'     \item A vertical reference line at x=1 (dashed)
#'     \item Appropriate labels and title
#'   }
#'
#' @examples
#' \dontrun{
#' df_observed <- data_observed(
#'   data = df_em,
#'   bias = "em",
#'   exposure = "Xstar",
#'   outcome = "Y",
#'   confounders = "C1"
#' )
#'
#' bp1 <- bias_params(coef_list = list(x = c(-2.10, 1.62, 0.63, 0.35)))
#' bp2 <- bias_params(coef_list = list(x = c(-2.10 * 2, 1.62 * 2, 0.63 * 2, 0.35 * 2)))
#'
#' result1 <- multibias_adjust(
#'   data_observed = df_observed,
#'   bias_params = bp1
#' )
#' result2 <- multibias_adjust(
#'   data_observed = df_observed,
#'   bias_params = bp2
#' )
#'
#' multibias_plot(
#'   data_observed = df_observed,
#'   multibias_result_list = list(
#'     "Adjusted with bias params" = result1,
#'     "Adjusted with bias params doubled" = result2
#'   )
#' )
#' }
#'
#' @export

multibias_plot <- function(
    data_observed,
    multibias_result_list,
    log_scale = FALSE) {
  summary_quietly <- purrr::quietly(summary)
  observed_summary <- summary_quietly(data_observed)$result

  observed_est <- observed_summary$estimate[2]
  observed_ci_low <- observed_summary$conf.low[2]
  observed_ci_high <- observed_summary$conf.high[2]

  df_observed <- tibble(
    type = "Observed",
    est = observed_est,
    ci_low = observed_ci_low,
    ci_high = observed_ci_high
  )

  list_names <- names(multibias_result_list)

  list_adjusted <- purrr::map(
    seq_along(multibias_result_list),
    function(list_index) {
      list_item <- multibias_result_list[[list_index]]
      name <- list_names[list_index]

      is_valid <- is.list(list_item) &&
        all(c("estimate", "ci") %in% names(list_item)) &&
        length(list_item$ci) == 2 &&
        is.numeric(list_item$estimate) &&
        is.numeric(list_item$ci)

      if (!is_valid) {
        warning(
          paste0(
            "Item ",
            name,
            " in multibias_result_list is not structured as expected. ",
            "Skipping this item."
          ),
          call. = FALSE
        )
        return(NULL)
      }
      tibble(
        type = name,
        est = list_item$estimate,
        ci_low = list_item$ci[1],
        ci_high = list_item$ci[2]
      )
    }
  )

  df_adjusted <- purrr::list_rbind(list_adjusted)
  df <- rbind(df_observed, as.data.frame(df_adjusted))

  # Set factor levels to control y-axis order
  df$type <- factor(df$type, levels = rev(c("Observed", list_names)))

  y_binary <- is_binary(data_observed$data[, data_observed$outcome])
  null_value <- if (y_binary) 1 else 0

  final <- ggplot2::ggplot(
    data = df, ggplot2::aes(x = .data$est, y = .data$type)
  ) +
    ggplot2::geom_vline(
      ggplot2::aes(xintercept = null_value), linetype = "dashed"
    ) +
    ggplot2::geom_point(color = "blue") +
    ggplot2::geom_errorbar(
      ggplot2::aes(xmin = .data$ci_low, xmax = .data$ci_high),
      linewidth = .5,
      width = .2,
      color = "gray50",
      orientation = "y"
    ) +
    ggplot2::labs(
      title = "Observed vs. Adjusted Effect Estimates",
      subtitle = paste0(
        "Sensitivity analysis of the ",
        data_observed$exposure,
        "-",
        data_observed$outcome,
        " relationship"
      ),
      x = "Effect Estimate",
      y = ""
    ) +
    ggplot2::theme_bw()

  if (log_scale) {
    if (any(df$est <= 0 | df$ci_low <= 0, na.rm = TRUE)) {
      warning(
        paste0(
          "Some estimates or lower CI bounds are zero or negative. ",
          "Log scale may produce NaNs, errors, or missing data points. ",
          "Review data or consider if log scale is appropriate."
        ),
        call. = FALSE
      )
    }
    final <- final +
      ggplot2::scale_x_log10()
  }

  return(final)
}

#' Represent observed causal data
#'
#' @description
#' `data_observed` combines the observed dataframe with specific identification
#' of the columns corresponding to the exposure, outcome, and confounders. It is
#' an essential input of the [multibias_adjust()] function.
#'
#' @param data Dataframe for bias analysis.
#' @param bias String type(s) of bias distorting the effect of the exposure
#' on the outcome. Can choose from a subset of the following: "uc", "em", "om",
#' "sel". These correspond to uncontrolled confounding, exposure
#' misclassification, outcome misclassification, and selection bias,
#' respectively.
#' @param exposure String name of the column in `data` corresponding to the
#' exposure variable.
#' @param outcome String name of the column in `data` corresponding to the
#' outcome variable.
#' @param confounders String name(s) of the column(s) in `data` corresponding
#' to the confounding variable(s).
#'
#' @return An object of class `data_observed` containing:
#'   \item{data}{A dataframe with the selected columns}
#'   \item{bias}{The type(s) of bias present}
#'   \item{exposure}{The name of the exposure variable}
#'   \item{outcome}{The name of the outcome variable}
#'   \item{confounders}{The name(s) of the confounder variable(s)}
#'
#' @examples
#' df <- data_observed(
#'   data = df_uc,
#'   bias = "uc",
#'   exposure = "X_bi",
#'   outcome = "Y_bi",
#'   confounders = c("C1", "C2", "C3")
#' )
#'
#' @export

data_observed <- function(
    data,
    bias,
    exposure,
    outcome,
    confounders = NULL) {
  stopifnot(
    is.data.frame(data),
    is.character(bias),
    is.character(exposure) & length(exposure) == 1,
    is.character(outcome) & length(outcome) == 1,
    is.character(confounders) | is.null(confounders)
  )

  required_vars <- c(exposure, outcome, confounders)
  required_vars <- required_vars[!is.null(required_vars)]

  if (!all(required_vars %in% names(data))) {
    missing_vars <- required_vars[!required_vars %in% names(data)]
    stop(
      paste0(
        "Input variable(s) missing in data: ",
        paste(missing_vars, collapse = ", ")
      )
    )
  }

  df <- data %>%
    select(all_of(c(exposure, outcome, confounders)))

  acceptable_biases <- c("uc", "em", "om", "sel")
  if (!all(bias %in% acceptable_biases)) {
    stop(
      paste0(
        "Unacceptable bias input. ",
        "Biases must include a subset of the following: ",
        "uc, em, om, sel"
      )
    )
  }

  obj <- structure(
    list(
      data = df,
      bias = bias,
      exposure = exposure,
      outcome = outcome,
      confounders = confounders
    ),
    class = "data_observed"
  )

  return(obj)
}


#' Print method for data_observed objects
#'
#' @description
#' Prints a formatted summary of a `data_observed` object, including:
#' - The types of biases present
#' - The exposure, outcome, and confounder variables
#' - A preview of the first 5 rows of data
#'
#' @param x A `data_observed` object
#' @param ... Additional arguments passed to print
#'
#' @return The input object invisibly, allowing for method chaining
#'
#' @method print data_observed
#' @keywords internal
#' @export

print.data_observed <- function(x, ...) {
  bias_map <- c(
    "uc" = "Uncontrolled Confounding",
    "em" = "Exposure Misclassification",
    "om" = "Outcome Misclassification",
    "sel" = "Selection Bias"
  )
  bias_description <- bias_map[x$bias]

  cat("Observed Data\n")
  cat("---------------------------------\n")
  cat("The following biases are present:", "\n")
  cat(paste(bias_description, collapse = "\n"), "\n")
  cat("---------------------------------\n")
  cat("Exposure:", x$exposure, "\n")
  cat("Outcome:", x$outcome, "\n")
  if (!is.null(x$confounders)) {
    cat("Confounders:", paste(x$confounders, collapse = ", "), "\n")
  }
  cat("---------------------------------\n")
  cat("Data head: \n")
  print(x$data[seq_len(min(5, nrow(x$data))), ])
  invisible(x)
}


#' Summary method for data_observed objects
#'
#' @description
#' Provides a statistical summary of the observed data by fitting either:
#' - A logistic regression model for binary outcomes
#' - A linear regression model for continuous outcomes
#'
#' The model includes the exposure and all confounders as predictors.
#' For binary outcomes, estimates are exponentiated to show odds ratios.
#'
#' @param object A `data_observed` object
#' @param ... Additional arguments passed to summary
#'
#' @return A data frame containing model coefficients, standard errors,
#'         confidence intervals, and p-values. For binary outcomes,
#'         coefficients are exponentiated to show odds ratios.
#'
#' @method summary data_observed
#' @importFrom broom tidy
#' @keywords internal
#' @export

summary.data_observed <- function(object, ...) {
  df <- data.frame(
    exposure = object$data[, object$exposure],
    Y = object$data[, object$outcome]
  )
  names(df)[1] <- object$exposure

  df <- bind_cols(
    df,
    object$data %>%
      select(all_of(object$confounders))
  )

  if (all(df$Y %in% 0:1)) {
    y_binary <- TRUE
  } else {
    y_binary <- FALSE
  }

  if (y_binary) {
    mod <- glm(
      Y ~ .,
      family = binomial(link = "logit"),
      data = df
    )
    final <- broom::tidy(mod, conf.int = TRUE, exponentiate = TRUE)
    cat(
      "Note: Estimates are exponentiated (odds ratios) for binary outcomes\n\n"
    )
  } else {
    mod <- lm(
      Y ~ .,
      data = df
    )
    final <- broom::tidy(mod, conf.int = TRUE, exponentiate = FALSE)
    cat("Note: Estimates are not exponentiated for continuous outcomes\n\n")
  }

  return(final)
}


#' Represent validation causal data
#'
#' @description
#' `data_validation` is one of two different options to represent bias
#' assumptions for bias adjustment. It combines the validation dataframe
#' with specific identification of the appropriate columns for bias adjustment,
#' including: true exposure, true outcome, confounders, misclassified exposure,
#' misclassified outcome, and selection. The purpose of validation data is to
#' use an external data source to transport the necessary causal relationships
#' that are missing in the observed data.
#'
#' @param data Dataframe of validation data
#' @param true_exposure String name of the column in `data` corresponding to
#' the true exposure.
#' @param true_outcome String name of the column in `data` corresponding to
#' the true outcome.
#' @param confounders String name(s) of the column(s) in `data` corresponding
#' to the confounding variable(s).
#' @param misclassified_exposure String name of the column in `data`
#' corresponding to the misclassified exposure.
#' @param misclassified_outcome String name of the column in `data`
#' corresponding to the misclassified outcome.
#' @param selection String name of the column in `data` corresponding to the
#' selection indicator.
#'
#' @return An object of class `data_validation` containing:
#'   \item{data}{A dataframe with the selected columns}
#'   \item{true_exposure}{The name of the true exposure variable}
#'   \item{true_outcome}{The name of the true outcome variable}
#'   \item{confounders}{The name(s) of the confounder variable(s)}
#'   \item{misclassified_exposure}{The name of the misclassified exposure variable}
#'   \item{misclassified_outcome}{The name of the misclassified outcome variable}
#'   \item{selection}{The name of the selection indicator variable}
#'
#' @examples
#' df <- data_validation(
#'   data = df_sel_source,
#'   true_exposure = "X",
#'   true_outcome = "Y",
#'   confounders = c("C1", "C2", "C3"),
#'   selection = "S"
#' )
#'
#' @export

data_validation <- function(
    data,
    true_exposure,
    true_outcome,
    confounders = NULL,
    misclassified_exposure = NULL,
    misclassified_outcome = NULL,
    selection = NULL) {
  stopifnot(
    is.data.frame(data),
    is.character(true_exposure) & length(true_exposure) == 1,
    is.character(true_outcome) & length(true_outcome) == 1,
    is.character(confounders) | is.null(confounders),
    (is.character(misclassified_exposure) &
      length(misclassified_exposure) == 1) |
      is.null(misclassified_exposure),
    (is.character(misclassified_outcome) &
      length(misclassified_outcome) == 1) |
      is.null(misclassified_outcome),
    (is.character(selection) & length(selection) == 1) | is.null(selection)
  )

  required_vars <- c(
    true_exposure, true_outcome,
    confounders, misclassified_exposure,
    misclassified_outcome, selection
  )
  required_vars <- required_vars[!is.null(required_vars)]

  if (!all(required_vars %in% names(data))) {
    missing_vars <- required_vars[!required_vars %in% names(data)]
    stop(
      paste0(
        "Input variable(s) missing in data: ",
        paste(missing_vars, collapse = ", ")
      )
    )
  }

  cols <- c(
    true_exposure,
    true_outcome,
    confounders,
    misclassified_exposure,
    misclassified_outcome,
    selection
  )

  df <- data %>%
    select(all_of(cols))

  obj <- list(
    data = df,
    true_exposure = true_exposure,
    true_outcome = true_outcome,
    confounders = confounders,
    misclassified_exposure = misclassified_exposure,
    misclassified_outcome = misclassified_outcome,
    selection = selection
  )

  class(obj) <- "data_validation"
  return(obj)
}


#' Print method for data_validation objects
#'
#' @description
#' Prints a formatted summary of a `data_validation` object, including:
#' - The true exposure and outcome variables
#' - Any confounders, misclassified variables, or selection indicators
#' - A preview of the first 5 rows of data
#'
#' @param x A `data_validation` object
#' @param ... Additional arguments passed to print
#'
#' @return The input object invisibly
#'
#' @keywords internal
#' @export

print.data_validation <- function(x, ...) {
  cat("Validation Data\n")
  cat("------------------\n")
  cat("True exposure:", x$true_exposure, "\n")
  cat("True outcome:", x$true_outcome, "\n")
  if (!is.null(x$confounders)) {
    cat("Confounders:", paste(x$confounders, collapse = ", "), "\n")
  }
  if (!is.null(x$misclassified_exposure)) {
    cat("Misclassified exposure:", x$misclassified_exposure, "\n")
  }
  if (!is.null(x$misclassified_outcome)) {
    cat("Misclassified outcome:", x$misclassified_outcome, "\n")
  }
  if (!is.null(x$selection)) {
    cat("Selection:", x$selection, "\n")
  }
  cat("Data head: \n")
  print(x$data[seq_len(min(5, nrow(x$data))), ])
  invisible(x)
}

set.seed(1234)
nreps <- 5

# SEPARATE BIAS PARAMETERS
# 0 confounders

# nobias_model <- glm(
#   Y ~ X + U,
#   family = binomial(link = "logit"),
#   data = df_uc_om_source
# )
# or_true <- exp(summary(nobias_model)$coef[2, 1])

# u_model <- glm(U ~ X + Y,
#                family = binomial(link = "logit"),
#                data = df_uc_om_source)
# y_model <- glm(Y ~ X + Ystar,
#                family = binomial(link = "logit"),
#                data = df_uc_om_source)

# df_observed <- data_observed(
#   df_uc_om,
#   bias = c("uc", "om"),
#   exposure = "X",
#   outcome = "Ystar",
#   confounders = NULL
# )
# list_for_uc_om <- list(
#   u = as.vector(coef(u_model)),
#   y = as.vector(coef(y_model))
# )
# bp_uc_om <- bias_params(coef_list = list_for_uc_om)

# single_run <- multibias_adjust(
#   df_observed,
#   bias_params = bp_uc_om
# )

# bs_run <- multibias_adjust(
#   df_observed,
#   bias_params = bp_uc_om,
#   bootstrap = TRUE,
#   bootstrap_reps = nreps
# )

# test_that("odds ratio and confidence interval output", {
#   expect_gt(bs_run$estimate, or_true - 0.1)
#   expect_lt(bs_run$estimate, or_true + 0.1)
#   expect_vector(
#     single_run$ci,
#     ptype = double(),
#     size = 2
#   )
# })

# 3 confounders

nobias_model <- glm(
  Y ~ X + C1 + C2 + C3 + U,
  family = binomial(link = "logit"),
  data = df_uc_om_source
)
or_true <- exp(summary(nobias_model)$coef[2, 1])

u_model <- glm(U ~ X + Y,
               family = binomial(link = "logit"),
               data = df_uc_om_source)
y_model <- glm(Y ~ X + Ystar + C1 + C2 + C3,
               family = binomial(link = "logit"),
               data = df_uc_om_source)

df_observed <- data_observed(
  df_uc_om,
  bias = c("uc", "om"),
  exposure = "X",
  outcome = "Ystar",
  confounders = c("C1", "C2", "C3")
)
list_for_uc_om <- list(
  u = as.vector(coef(u_model)),
  y = as.vector(coef(y_model))
)
bp_uc_om <- bias_params(coef_list = list_for_uc_om)

single_run <- multibias_adjust(
  df_observed,
  bias_params = bp_uc_om
)

bs_run <- multibias_adjust(
  df_observed,
  bias_params = bp_uc_om,
  bootstrap = TRUE,
  bootstrap_reps = nreps
)

test_that("odds ratio and confidence interval output", {
  expect_gt(bs_run$estimate, or_true - 0.1)
  expect_lt(bs_run$estimate, or_true + 0.1)
  expect_vector(
    single_run$ci,
    ptype = double(),
    size = 2
  )
})

# MULTINOMIAL BIAS PARAMETERS
# 0 confounders

nobias_model <- glm(
  Y ~ X + U,
  family = binomial(link = "logit"),
  data = df_uc_om_source
)
or_true <- exp(summary(nobias_model)$coef[2, 1])

# library(nnet)
# uy_model <- multinom(
#   paste0(U, Y) ~ X + Ystar,
#   data = df_uc_om_source
# )
# summary(uy_model, digits = 2)

df_observed <- data_observed(
  df_uc_om,
  bias = c("uc", "om"),
  exposure = "X",
  outcome = "Ystar",
  confounders = NULL
)
list_for_uc_om <- list(
  u1y0 = c(-0.32, 0.59, 0.01),
  u0y1 = c(-2.98, 0.71, 1.65),
  u1y1 = c(-2.59, 1.27, 1.64)

)
bp_uc_om <- bias_params(coef_list = list_for_uc_om)

single_run <- multibias_adjust(
  df_observed,
  bias_params = bp_uc_om
)

bs_run <- multibias_adjust(
  df_observed,
  bias_params = bp_uc_om,
  bootstrap = TRUE,
  bootstrap_reps = nreps
)

test_that("odds ratio and confidence interval output", {
  expect_gt(bs_run$estimate, or_true - 0.1)
  expect_lt(bs_run$estimate, or_true + 0.1)
  expect_vector(
    single_run$ci,
    ptype = double(),
    size = 2
  )
})

# 3 confounders

nobias_model <- glm(
  Y ~ X + C1 + C2 + C3 + U,
  family = binomial(link = "logit"),
  data = df_uc_om_source
)
or_true <- exp(summary(nobias_model)$coef[2, 1])

# uy_model <- multinom(
#   paste0(U, Y) ~ X + Ystar + C1 + C2 + C3,
#   data = df_uc_om_source
# )
# summary(uy_model, digits = 2)

df_observed <- data_observed(
  df_uc_om,
  bias = c("uc", "om"),
  exposure = "X",
  outcome = "Ystar",
  confounders = c("C1", "C2", "C3")
)
list_for_uc_om <- list(
  u1y0 = c(-0.2, 0.62, 0.01, -0.08, 0.10, -0.15),
  u0y1 = c(-3.3, 0.63, 1.65, 0.42, -0.85, 0.26),
  u1y1 = c(-2.7, 1.22, 1.64, 0.32, -0.77, 0.09)

)
bp_uc_om <- bias_params(coef_list = list_for_uc_om)

single_run <- multibias_adjust(
  df_observed,
  bias_params = bp_uc_om
)

bs_run <- multibias_adjust(
  df_observed,
  bias_params = bp_uc_om,
  bootstrap = TRUE,
  bootstrap_reps = nreps
)

test_that("odds ratio and confidence interval output", {
  expect_gt(bs_run$estimate, or_true - 0.1)
  expect_lt(bs_run$estimate, or_true + 0.1)
  expect_vector(
    single_run$ci,
    ptype = double(),
    size = 2
  )
})

# adjust with validation data

val_run <- multibias_adjust(
  data_observed = data_observed(
    df_uc_om,
    bias = c("uc", "om"),
    exposure = "X",
    outcome = "Ystar",
    confounders = c("C1", "C2", "C3")
  ),
  data_validation = data_validation(
    df_uc_om_source,
    true_exposure = "X",
    true_outcome = "Y",
    confounders = c("C1", "C2", "C3", "U"),
    misclassified_outcome = "Ystar"
  ),
  bootstrap = TRUE,
  bootstrap_reps = nreps
)

test_that("adjust_uc_om, validation data", {
  expect_gt(val_run$estimate, or_true - 0.1)
  expect_lt(val_run$estimate, or_true + 0.1)
})

library(vdiffr)
set.seed(1234)
nreps <- 5

# adjust with coefs
# 0 confounders

# nobias_model <- glm(
#   Y_bi ~ X_bi + U,
#   family = binomial(link = "logit"),
#   data = df_uc_source
# )
# or_true <- exp(summary(nobias_model)$coef[2, 1])

# u_model <- glm(
#   U ~ X_bi + Y_bi,
#   family = binomial(link = "logit"),
#   data = df_uc_source
# )

# df_observed <- data_observed(
#   df_uc,
#   bias = "uc",
#   exposure = "X_bi",
#   outcome = "Y_bi",
#   confounders = NULL
# )
# list_for_uc <- list(u = as.vector(coef(u_model)))
# bp_uc <- bias_params(coef_list = list_for_uc)

# single_run <- multibias_adjust(
#   df_observed,
#   bias_params = bp_uc
# )

# bs_run <- multibias_adjust(
#   df_observed,
#   bias_params = bp_uc,
#   bootstrap = TRUE,
#   bootstrap_reps = nreps
# )

# test_that("adjust_uc, 0 confounders: OR and CI output", {
#   expect_gt(bs_run$estimate, or_true - 0.1)
#   expect_lt(bs_run$estimate, or_true + 0.1)
#   expect_vector(
#     single_run$ci,
#     ptype = double(),
#     size = 2
#   )
# })

# 3 confounders

nobias_model <- lm(Y_cont ~ X_cont + C1 + C2 + C3 + U,
  data = df_uc_source
)
or_true <- summary(nobias_model)$coef[2, 1]

u_model <- glm(U ~ X_cont + Y_cont + C1 + C2 + C3,
  family = binomial(link = "logit"),
  data = df_uc_source
)

df_observed <- data_observed(
  df_uc,
  bias = "uc",
  exposure = "X_cont",
  outcome = "Y_cont",
  confounders = c("C1", "C2", "C3")
)
list_for_uc <- list(u = as.vector(coef(u_model)))
bp_uc <- bias_params(coef_list = list_for_uc)

single_run <- multibias_adjust(
  df_observed,
  bias_params = bp_uc
)

bs_run <- multibias_adjust(
  df_observed,
  bias_params = bp_uc,
  bootstrap = TRUE,
  bootstrap_reps = nreps
)

test_that("adjust_uc, 3 confounders: OR and CI output", {
  expect_gt(bs_run$estimate, or_true - 0.1)
  expect_lt(bs_run$estimate, or_true + 0.1)
  expect_vector(
    single_run$ci,
    ptype = double(),
    size = 2
  )
})

# adjust with validation data

val_run <- multibias_adjust(
  data_observed = data_observed(
    df_uc,
    bias = "uc",
    exposure = "X_cont",
    outcome = "Y_cont",
    confounders = c("C1", "C2", "C3")
  ),
  data_validation = data_validation(
    df_uc_source,
    true_exposure = "X_cont",
    true_outcome = "Y_cont",
    confounders = c("C1", "C2", "C3", "U")
  ),
  bootstrap = TRUE,
  bootstrap_reps = nreps
)

test_that("adjust_uc, validation data", {
  expect_gt(val_run$estimate, or_true - 0.1)
  expect_lt(val_run$estimate, or_true + 0.1)
})

# test plot

# plot <- multibias_plot(
#   data_observed = df_observed,
#   multibias_result_list = list("Adjusted" = val_run)
# )

# test_that("Plot output", {
#   skip_on_cran() # Skip test for CRAN submission
#   expect_doppelganger("multibias_plot", plot)
# })

set.seed(1234)
nreps <- 5

# SEPARATE BIAS PARAMETERS
# 0 confounders

# nobias_model <- glm(
#   Y ~ X + U,
#   family = binomial(link = "logit"),
#   data = df_uc_em_sel_source
# )
# or_true <- exp(summary(nobias_model)$coef[2, 1])

# u_model <- glm(
#   U ~ X + Y,
#   family = binomial(link = "logit"),
#   data = df_uc_em_sel_source
# )
# x_model <- glm(
#   X ~ Xstar + Y,
#   family = binomial(link = "logit"),
#   data = df_uc_em_sel_source
# )
# s_model <- glm(
#   S ~ Xstar + Y,
#   family = binomial(link = "logit"),
#   data = df_uc_em_sel_source
# )

# df_observed <- data_observed(
#   df_uc_em_sel,
#   bias = c("uc", "em", "sel"),
#   exposure = "Xstar",
#   outcome = "Y",
#   confounders = NULL
# )
# list_for_uc_em_sel <- list(
#   u = as.vector(coef(u_model)),
#   x = as.vector(coef(x_model)),
#   s = as.vector(coef(s_model))
# )
# bp_uc_em_sel <- bias_params(coef_list = list_for_uc_em_sel)

# single_run <- multibias_adjust(
#   df_observed,
#   bias_params = bp_uc_em_sel
# )

# bs_run <- multibias_adjust(
#   df_observed,
#   bias_params = bp_uc_em_sel,
#   bootstrap = TRUE,
#   bootstrap_reps = nreps
# )

# test_that("0 confounders: odds ratio and confidence interval output", {
#   expect_gt(bs_run$estimate, or_true - 0.1)
#   expect_lt(bs_run$estimate, or_true + 0.1)
#   expect_vector(
#     single_run$ci,
#     ptype = double(),
#     size = 2
#   )
# })

# 3 confounders

nobias_model <- glm(
  Y ~ X + C1 + C2 + C3 + U,
  family = binomial(link = "logit"),
  data = df_uc_em_sel_source
)
or_true <- exp(summary(nobias_model)$coef[2, 1])

u_model <- glm(
  U ~ X + Y,
  family = binomial(link = "logit"),
  data = df_uc_em_sel_source
)
x_model <- glm(
  X ~ Xstar + Y + C1 + C2 + C3,
  family = binomial(link = "logit"),
  data = df_uc_em_sel_source
)
s_model <- glm(
  S ~ Xstar + Y + C1 + C2 + C3,
  family = binomial(link = "logit"),
  data = df_uc_em_sel_source
)

df_observed <- data_observed(
  df_uc_em_sel,
  bias = c("uc", "em", "sel"),
  exposure = "Xstar",
  outcome = "Y",
  confounders = c("C1", "C2", "C3")
)
list_for_uc_em_sel <- list(
  u = as.vector(coef(u_model)),
  x = as.vector(coef(x_model)),
  s = as.vector(coef(s_model))
)
bp_uc_em_sel <- bias_params(coef_list = list_for_uc_em_sel)

single_run <- multibias_adjust(
  df_observed,
  bias_params = bp_uc_em_sel
)

bs_run <- multibias_adjust(
  df_observed,
  bias_params = bp_uc_em_sel,
  bootstrap = TRUE,
  bootstrap_reps = nreps
)

test_that("3 confounders: odds ratio and confidence interval output", {
  expect_gt(bs_run$estimate, or_true - 0.1)
  expect_lt(bs_run$estimate, or_true + 0.1)
  expect_vector(
    single_run$ci,
    ptype = double(),
    size = 2
  )
})

# MULTINOMIAL BIAS PARAMETERS
# 0 confounders

nobias_model <- glm(
  Y ~ X + U,
  family = binomial(link = "logit"),
  data = df_uc_em_sel_source
)
or_true <- exp(summary(nobias_model)$coef[2, 1])

# library(nnet)
# xu_model <- multinom(
#   paste(X, U) ~ Xstar + Y,
#   data = df_uc_em_sel_source
# )

s_model <- glm(
  S ~ Xstar + Y,
  family = binomial(link = "logit"),
  data = df_uc_em_sel_source
)

df_observed <- data_observed(
  df_uc_em_sel,
  bias = c("uc", "em", "sel"),
  exposure = "Xstar",
  outcome = "Y",
  confounders = NULL
)
list_for_uc_em_sel <- list(
  x1u0 = c(-1.92, 1.62, 0.70),
  x0u1 = c(-0.30, -0.01, 0.69),
  x1u1 = c(-1.64, 1.62, 1.35),
  s = as.vector(coef(s_model))
)
bp_uc_em_sel <- bias_params(coef_list = list_for_uc_em_sel)

single_run <- multibias_adjust(
  df_observed,
  bias_params = bp_uc_em_sel
)

bs_run <- multibias_adjust(
  df_observed,
  bias_params = bp_uc_em_sel,
  bootstrap = TRUE,
  bootstrap_reps = nreps
)

test_that("0 confounders: odds ratio and confidence interval output", {
  expect_gt(bs_run$estimate, or_true - 0.1)
  expect_lt(bs_run$estimate, or_true + 0.1)
  expect_vector(
    single_run$ci,
    ptype = double(),
    size = 2
  )
})

# 3 confounders

nobias_model <- glm(
  Y ~ X + C1 + C2 + C3 + U,
  family = binomial(link = "logit"),
  data = df_uc_em_sel_source
)
or_true <- exp(summary(nobias_model)$coef[2, 1])

# xu_model <- multinom(
#   paste(X, U) ~ Xstar + Y + C1 + C2 + C3,
#   data = df_uc_em_sel_source
# )

s_model <- glm(
  S ~ Xstar + Y + C1 + C2 + C3,
  family = binomial(link = "logit"),
  data = df_uc_em_sel_source
)

df_observed <- data_observed(
  df_uc_em_sel,
  bias = c("uc", "em", "sel"),
  exposure = "Xstar",
  outcome = "Y",
  confounders = c("C1", "C2", "C3")
)
list_for_uc_em_sel <- list(
  x1u0 = c(-2.78, 1.62, 0.61, 0.36, -0.27, 0.88),
  x0u1 = c(-0.17, -0.01, 0.71, -0.08, 0.07, -0.15),
  x1u1 = c(-2.36, 1.62, 1.29, 0.25, -0.06, 0.74),
  s = as.vector(coef(s_model))
)
bp_uc_em_sel <- bias_params(coef_list = list_for_uc_em_sel)

single_run <- multibias_adjust(
  df_observed,
  bias_params = bp_uc_em_sel
)

bs_run <- multibias_adjust(
  df_observed,
  bias_params = bp_uc_em_sel,
  bootstrap = TRUE,
  bootstrap_reps = nreps
)

test_that("3 confounders: odds ratio and confidence interval output", {
  expect_gt(bs_run$estimate, or_true - 0.1)
  expect_lt(bs_run$estimate, or_true + 0.1)
  expect_vector(
    single_run$ci,
    ptype = double(),
    size = 2
  )
})

# adjust with validation data

val_run <- multibias_adjust(
  data_observed = data_observed(
    df_uc_em_sel,
    bias = c("uc", "em", "sel"),
    exposure = "Xstar",
    outcome = "Y",
    confounders = c("C1", "C2", "C3")
  ),
  data_validation = data_validation(
    df_uc_em_sel_source,
    true_exposure = "X",
    true_outcome = "Y",
    confounders = c("C1", "C2", "C3", "U"),
    misclassified_exposure = "Xstar",
    selection = "S"
  ),
  bootstrap = TRUE,
  bootstrap_reps = nreps
)

test_that("adjust_uc_em_sel, validation data", {
  expect_gt(val_run$estimate, or_true - 0.1)
  expect_lt(val_run$estimate, or_true + 0.1)
})

# Adjust for selection bias

# the following functions feed into adjust_sel():
# adjust_sel_val() (data_validation input, method: weighting),
# adjust_sel_coef() (bias_params input, method: weighting)

adjust_sel_val <- function(
    data_observed,
    data_validation) {
  if (!all(data_observed$confounders %in% data_validation$confounders)) {
    stop(
      "All confounders in observed data must be present in validation data.",
      call. = FALSE
    )
  }

  if (is.null(data_validation$selection)) {
    stop(
      paste0(
        "Attempting to adjust for selection bias.",
        "\n",
        "Validation data must have a selection indicator column specified."
      ),
      call. = FALSE
    )
  }

  n <- nrow(data_observed$data)

  df <- data.frame(
    X = data_observed$data[, data_observed$exposure],
    Y = data_observed$data[, data_observed$outcome]
  )
  df <- bind_cols(
    df,
    data_observed$data %>%
      select(all_of(data_observed$confounders))
  )

  if (all(df$Y %in% 0:1)) {
    y_binary <- TRUE
  } else {
    y_binary <- FALSE
  }

  df_val <- data.frame(
    X = data_validation$data[, data_validation$true_exposure],
    Y = data_validation$data[, data_validation$true_outcome],
    S = data_validation$data[, data_validation$selection]
  )

  force_match(
    df$X,
    df_val$X,
    "Exposures from both datasets must both be binary or both be continuous."
  )
  force_match(
    df$Y,
    df_val$Y,
    "Outcomes from both datasets must both be binary or both be continuous."
  )
  force_binary(
    df_val$S,
    "Selection indicator in validation data must be a binary integer."
  )

  s_mod <- glm(S ~ X + Y,
    family = binomial(link = "logit"),
    data = df_val
  )

  s_mod_coefs <- coef(s_mod)
  s_mod_se <- sqrt(diag(vcov(s_mod)))

  # Sample coefficients independently using their standard errors
  s_mod_coefs_sampled <- rnorm(length(s_mod_coefs), s_mod_coefs, s_mod_se)
  s_pred <- s_mod_coefs_sampled[1]

  for (i in 2:length(s_mod_coefs_sampled)) {
    var_name <- names(s_mod_coefs)[i]
    s_pred <- s_pred + df[[var_name]] * s_mod_coefs_sampled[i]
  }

  df$Spred <- plogis(s_pred)

  if (y_binary) {
    suppressWarnings({
      final <- glm(
        Y ~ X + . - Spred,
        family = binomial(link = "logit"),
        weights = (1 / df$Spred),
        data = df
      )
    })
  } else {
    suppressWarnings({
      final <- lm(
        Y ~ X + . - Spred,
        weights = (1 / df$Spred),
        data = df
      )
    })
  }

  return(final)
}


adjust_sel_coef <- function(
    data_observed,
    s_model_coefs) {
  data <- data_observed$data
  n <- nrow(data)
  confounders <- data_observed$confounders
  len_c <- length(confounders)
  len_s_coefs <- length(s_model_coefs)

  x <- data[, data_observed$exposure]
  y <- data[, data_observed$outcome]

  force_len(
    len_s_coefs,
    3,
    paste0(
      "Incorrect length of S model coefficients. ",
      "Length should equal 3."
    )
  )

  if (all(y %in% 0:1)) {
    y_binary <- TRUE
  } else {
    y_binary <- FALSE
  }

  # Extract S model coefficients
  s1_0 <- s_model_coefs[1]
  s1_x <- s_model_coefs[2]
  s1_y <- s_model_coefs[3]

  # Create base dataframe
  df <- data.frame(X = x, Y = y)

  # Add confounders if they exist
  if (!is.null(confounders)) {
    for (i in seq_along(confounders)) {
      df[[paste0("C", i)]] <- data[, confounders[i]]
    }
  }

  # Calculate selection probabilities
  df$pS <- plogis(s1_0 + s1_x * df$X + s1_y * df$Y)

  # Construct final model formula
  model_terms <- c("X")
  if (!is.null(confounders)) {
    model_terms <- c(model_terms, paste0("C", seq_along(confounders)))
  }
  model_formula <- as.formula(
    paste("Y ~", paste(model_terms, collapse = " + "))
  )

  # Fit final model with weights
  suppressWarnings({
    if (y_binary) {
      final <- glm(
        model_formula,
        family = binomial(link = "logit"),
        weights = (1 / df$pS),
        data = df
      )
    } else {
      final <- lm(
        model_formula,
        weights = (1 / df$pS),
        data = df
      )
    }
  })

  return(final)
}


adjust_sel <- function(
    data_observed,
    data_validation = NULL,
    bias_params = NULL,
    level = 0.95) {
  data <- data_observed$data
  x <- data[, data_observed$exposure]
  y <- data[, data_observed$outcome]

  if (all(y %in% 0:1)) {
    y_binary <- TRUE
  } else {
    y_binary <- FALSE
  }

  if (!is.null(data_validation)) {
    final <- adjust_sel_val(
      data_observed,
      data_validation
    )
  } else if (!is.null(bias_params)) {
    if (is.null(bias_params$coef_list$s)) {
      stop(
        "bias_params must specify parameters for s to adjust for selection bias",
        call. = FALSE
      )
    }
    final <- adjust_sel_coef(
      data_observed,
      s_model_coefs = bias_params$coef_list$s
    )
  }

  calculate_results(final, level, y_binary)
}

#' Simulated data with uncontrolled confounding
#'
#' Data containing one source of bias, three known confounders, and
#' 100,000 observations. This data is obtained from `df_uc_source`
#' by removing the column *U*. The resulting data corresponds to
#' what a researcher would see in the real-world: information on known
#' confounders (*C1*, *C2*, and *C3*), but not for
#' confounder *U*.
#' As seen in `df_uc_source`, the true, unbiased exposure-outcome
#' effect estimate = 2.
#'
#' @format A dataframe with 100,000 rows and 7 columns:
#' \describe{
#'     \item{X_bi}{binary exposure, 1 = present and 0 = absent}
#'     \item{X_cont}{continuous exposure}
#'     \item{Y_bi}{binary outcome corresponding to exposure *X_bi*, 1 = present and 0 = absent}
#'     \item{Y_cont}{continuous outcome corresponding to exposure *X_cont*}
#'     \item{C1}{1st confounder, 1 = present and 0 = absent}
#'     \item{C2}{2nd confounder, 1 = present and 0 = absent}
#'     \item{C3}{3rd confounder, 1 = present and 0 = absent}
#' }
"df_uc"

#' Data source for `df_uc`
#'
#' Data with complete information on one source of bias, three known
#' confounders, and 100,000 observations. This data is used to derive
#' `df_uc` and can be used to obtain bias parameters for purposes
#' of validating the simultaneous multi-bias adjustment method with
#' `df_uc`. With this source data, the fitted regression
#' \ifelse{html}{\out{g(P(Y=1)) = &alpha;<sub>0</sub> + &alpha;<sub>1</sub>X + &alpha;<sub>2</sub>C1 + &alpha;<sub>3</sub>C2 + &alpha;<sub>4</sub>C3 + &alpha;<sub>5</sub>U}}{\eqn{logit(P(Y=1)) = \alpha_0 + \alpha_1 X + \alpha_2 C1 + \alpha_3 C2 + \alpha_4 C3 + \alpha_5 U}}
#' shows that the true, unbiased exposure-outcome effect estimate = 2 when:
#' \enumerate{
#'   \item g = logit, Y = *Y_bi*, and X = *X_bi* or
#'   \item g = identity, Y = *Y_cont*, X = *X_cont*.
#' }
#'
#' @format A dataframe with 100,000 rows and 8 columns:
#' \describe{
#'     \item{X_bi}{binary exposure, 1 = present and 0 = absent}
#'     \item{X_cont}{continuous exposure}
#'     \item{Y_bi}{binary outcome corresponding to exposure *X_bi*, 1 = present and 0 = absent}
#'     \item{Y_cont}{continuous outcome corresponding to exposure *X_cont*}
#'     \item{C1}{1st confounder, 1 = present and 0 = absent}
#'     \item{C2}{2nd confounder, 1 = present and 0 = absent}
#'     \item{C3}{3rd confounder, 1 = present and 0 = absent}
#'     \item{U}{uncontrolled confounder, 1 = present and 0 = absent}
#' }
"df_uc_source"

#' Represent bias parameters
#'
#' @description
#' `bias_params` is one of two different options to represent bias assumptions for bias adjustment. The [multibias_adjust()] function will apply the assumptions from these models and use them to adjust for biases in the observed data. It takes one input, a list, where each item in the list corresponds to the necessary models for bias adjustment. See below for bias models.
#'
#' For each of the following bias models, the variables are defined:
#' \itemize{
#'   \item X = True exposure
#'   \item X* = Misclassified exposure
#'   \item Y = True outcome
#'   \item Y* = Misclassified outcome
#'   \item C = Known confounder
#'   \item j = Number of known confounders
#'   \item U = Uncontrolled confounder
#'   \item S = Selection indicator
#' }
#'
#' \describe{
#'   \item{Uncontrolled confounding}{
#'     \ifelse{html}{\out{logit(P(U=1)) = &alpha;<sub>0</sub> + &alpha;<sub>1</sub>X + &alpha;<sub>2</sub>Y + &alpha;<sub>2+j</sub>C<sub>j</sub> }}{\eqn{logit(P(U=1)) = \alpha_0 + \alpha_1 X + \alpha_2 Y + \alpha_{2+j} C_j }}
#'   }
#'
#'   \item{Exposure misclassification}{
#'     \ifelse{html}{\out{logit(P(X=1)) = &delta;<sub>0</sub> + &delta;<sub>1</sub>X* + &delta;<sub>2</sub>Y + &delta;<sub>2+j</sub>C<sub>j</sub> }}{\eqn{logit(P(X=1)) = \delta_0 + \delta_1 X^* + \delta_2 Y + \delta_{2+j} C_j }}
#'   }
#'
#'   \item{Outcome misclassification}{
#'     \ifelse{html}{\out{logit(P(Y=1)) = &delta;<sub>0</sub> + &delta;<sub>1</sub>X + &delta;<sub>2</sub>Y* + &delta;<sub>2+j</sub>C<sub>j</sub> }}{\eqn{logit(P(Y=1)) = \delta_0 + \delta_1 X + \delta_2 Y^* + \delta_{2+j} C_j }}
#'   }
#'
#'   \item{Selection bias}{
#'     \ifelse{html}{\out{logit(P(S=1)) = &beta;<sub>0</sub> + &beta;<sub>1</sub>X + &beta;<sub>2</sub>Y }}{\eqn{logit(P(S=1)) = \beta_0 + \beta_1 X + \beta_2 Y }}
#'   }
#'
#'   \item{Uncontrolled Confounding & Exposure Misclassification (Option 1)}{
#'     \ifelse{html}{\out{logit(P(U=1)) = &alpha;<sub>0</sub> + &alpha;<sub>1</sub>X + &alpha;<sub>2</sub>Y <br> logit(P(X=1)) = &delta;<sub>0</sub> + &delta;<sub>1</sub>X* + &delta;<sub>2</sub>Y + &delta;<sub>2+j</sub>C<sub>j</sub> }}{\eqn{logit(P(U=1)) = \alpha_0 + \alpha_1 X + \alpha_2 Y \\ logit(P(X=1)) = \delta_0 + \delta_1 X^* + \delta_2 Y + \delta_{2+j} C_j }}
#'   }
#'
#'   \item{Uncontrolled Confounding & Exposure Misclassification (Option 2)}{
#'     \ifelse{html}{\out{log(P(X=1,U=0)/P(X=0,U=0)) = &gamma;<sub>1,0</sub> + &gamma;<sub>1,1</sub>X* + &gamma;<sub>1,2</sub>Y + &gamma;<sub>1,2+j</sub>C<sub>j</sub> <br> log(P(X=0,U=1)/P(X=0,U=0)) = &gamma;<sub>2,0</sub> + &gamma;<sub>2,1</sub>X* + &gamma;<sub>2,2</sub>Y + &gamma;<sub>2,2+j</sub>C<sub>j</sub> <br> log(P(X=1,U=1)/P(X=0,U=0)) = &gamma;<sub>3,0</sub> + &gamma;<sub>3,1</sub>X* + &gamma;<sub>3,2</sub>Y + &gamma;<sub>3,2+j</sub>C<sub>j</sub> }}{\eqn{log(P(X=1,U=0)/P(X=0,U=0)) = \gamma_{1,0} + \gamma_{1,1} X^* + \gamma_{1,2} Y + \gamma_{1,2+j} C_j \\ log(P(X=0,U=1)/P(X=0,U=0)) = \gamma_{2,0} + \gamma_{2,1} X^* + \gamma_{2,2} Y + \gamma_{2,2+j} C_j \\ log(P(X=1,U=1)/P(X=0,U=0)) = \gamma_{3,0} + \gamma_{3,1} X^* + \gamma_{3,2} Y + \gamma_{3,2+j} C_j }}
#'   }
#'
#'   \item{Uncontrolled Confounding & Outcome Misclassification (Option 1)}{
#'     \ifelse{html}{\out{logit(P(U=1)) = &alpha;<sub>0</sub> + &alpha;<sub>1</sub>X + &alpha;<sub>2</sub>Y <br> logit(P(Y=1)) = &delta;<sub>0</sub> + &delta;<sub>1</sub>X + &delta;<sub>2</sub>Y* + &delta;<sub>2+j</sub>C<sub>j</sub> }}{\eqn{logit(P(U=1)) = \alpha_0 + \alpha_1 X + \alpha_2 Y \\ logit(P(Y=1)) = \delta_0 + \delta_1 X + \delta_2 Y^* + \delta_{2+j} C_j }}
#'   }
#'
#'   \item{Uncontrolled Confounding & Outcome Misclassification (Option 2)}{
#'     \ifelse{html}{\out{log(P(U=1,Y=0)/P(U=0,Y=0)) = &gamma;<sub>1,0</sub> + &gamma;<sub>1,1</sub>X + &gamma;<sub>1,2</sub>Y* + &gamma;<sub>1,2+j</sub>C<sub>j</sub> <br> log(P(U=0,Y=1)/P(U=0,Y=0)) = &gamma;<sub>2,0</sub> + &gamma;<sub>2,1</sub>X + &gamma;<sub>2,2</sub>Y* + &gamma;<sub>2,2+j</sub>C<sub>j</sub> <br> log(P(U=1,Y=1)/P(U=0,Y=0)) = &gamma;<sub>3,0</sub> + &gamma;<sub>3,1</sub>X + &gamma;<sub>3,2</sub>Y* + &gamma;<sub>3,2+j</sub>C<sub>j</sub> }}{\eqn{log(P(U=1,Y=0)/P(U=0,Y=0)) = \gamma_{1,0} + \gamma_{1,1} X + \gamma_{1,2} Y^* + \gamma_{1,2+j} C_j \\ log(P(U=0,Y=1)/P(U=0,Y=0)) = \gamma_{2,0} + \gamma_{2,1} X + \gamma_{2,2} Y^* + \gamma_{2,2+j} C_j \\ log(P(U=1,Y=1)/P(U=0,Y=0)) = \gamma_{3,0} + \gamma_{3,1} X + \gamma_{3,2} Y^* + \gamma_{3,2+j} C_j }}
#'   }
#'
#'   \item{Uncontrolled Confounding & Selection Bias}{
#'     \ifelse{html}{\out{logit(P(U=1)) = &alpha;<sub>0</sub> + &alpha;<sub>1</sub>X + &alpha;<sub>2</sub>Y + &alpha;<sub>2+j</sub>C<sub>j</sub> <br> logit(P(S=1)) = &beta;<sub>0</sub> + &beta;<sub>1</sub>X + &beta;<sub>2</sub>Y }}{\eqn{logit(P(U=1)) = \alpha_0 + \alpha_1 X + \alpha_2 Y + \alpha_{2+j} C_j \\ logit(P(S=1)) = \beta_0 + \beta_1 X + \beta_2 Y }}
#'   }
#'
#'  \item{Exposure Misclassification & Outcome Misclassification (Option 1)}{
#'    \ifelse{html}{\out{logit(P(X=1)) = &delta;<sub>0</sub> + &delta;<sub>1</sub>X* + &delta;<sub>2</sub>Y* + &delta;<sub>2+j</sub>C<sub>j</sub> <br> logit(P(Y=1)) = &beta;<sub>0</sub> + &beta;<sub>1</sub>X + &beta;<sub>2</sub>Y* + &beta;<sub>2+j</sub>C<sub>j</sub> }}{\eqn{logit(P(X=1)) = \delta_0 + \delta_1 X^* + \delta_2 Y^* + \delta_{2+j} C_j \\ logit(P(Y=1)) = \beta_0 + \beta_1 X + \beta_2 Y^* + \beta_{2+j} C_j }}
#'  }
#'
#'   \item{Exposure Misclassification & Outcome Misclassification (Option 2)}{
#'     \ifelse{html}{\out{log(P(X=1,Y=0) / P(X=0,Y=0)) = &gamma;<sub>1,0</sub> + &gamma;<sub>1,1</sub>X* + &gamma;<sub>1,2</sub>Y* + &gamma;<sub>1,2+j</sub>C<sub>j</sub> <br> log(P(X=0,Y=1) / P(X=0,Y=0)) = &gamma;<sub>2,0</sub> + &gamma;<sub>2,1</sub>X* + &gamma;<sub>2,2</sub>Y* + &gamma;<sub>2,2+j</sub>C<sub>j</sub> <br> log(P(X=1,Y=1) / P(X=0,Y=0)) = &gamma;<sub>3,0</sub> + &gamma;<sub>3,1</sub>X* + &gamma;<sub>3,2</sub>Y* + &gamma;<sub>3,2+j</sub>C<sub>j</sub> }}{\eqn{log(P(X=1,Y=0) / P(X=0,Y=0)) = \gamma_{1,0} + \gamma_{1,1} X^* + \gamma_{1,2} Y^* + \gamma_{1,2+j} C_j \\ log(P(X=0,Y=1) / P(X=0,Y=0)) = \gamma_{2,0} + \gamma_{2,1} X^* + \gamma_{2,2} Y^* + \gamma_{2,2+j} C_j \\ log(P(X=1,Y=1) / P(X=0,Y=0)) = \gamma_{3,0} + \gamma_{3,1} X^* + \gamma_{3,2} Y^* + \gamma_{3,2+j} C_j }}
#'   }
#'
#'   \item{Exposure Misclassification & Selection Bias}{
#'     \ifelse{html}{\out{logit(P(X=1)) = &delta;<sub>0</sub> + &delta;<sub>1</sub>X* + &delta;<sub>2</sub>Y + &delta;<sub>2+j</sub>C<sub>j</sub> <br> logit(P(S=1)) = &beta;<sub>0</sub> + &beta;<sub>1</sub>X* + &beta;<sub>2</sub>Y + &beta;<sub>2+j</sub>C<sub>j</sub> }}{\eqn{logit(P(X=1)) = \delta_0 + \delta_1 X^* + \delta_2 Y + \delta_{2+j} C_j \\ logit(P(S=1)) = \beta_0 + \beta_1 X^* + \beta_2 Y + \beta_{2+j} C_j }}
#'   }
#'
#'  \item{Outcome Misclassification & Selection Bias}{
#'     \ifelse{html}{\out{logit(P(Y=1)) = &delta;<sub>0</sub> + &delta;<sub>1</sub>X + &delta;<sub>2</sub>Y* + &delta;<sub>2+j</sub>C<sub>j</sub> <br> logit(P(S=1)) = &beta;<sub>0</sub> + &beta;<sub>1</sub>X + &beta;<sub>2</sub>Y* + &beta;<sub>2+j</sub>C<sub>j</sub> }}{\eqn{logit(P(Y=1)) = \delta_0 + \delta_1 X + \delta_2 Y^* + \delta_{2+j} C_j \\ logit(P(S=1)) = \beta_0 + \beta_1 X + \beta_2 Y^* + \beta_{2+j} C_j }}
#'   }
#'   \item{Uncontrolled Confounding, Exposure Misclassification, and Selection Bias (Option 1)}{
#'     \ifelse{html}{\out{logit(P(U=1)) = &alpha;<sub>0</sub> + &alpha;<sub>1</sub>X + &alpha;<sub>2</sub>Y <br> logit(P(X=1)) = &delta;<sub>0</sub> + &delta;<sub>1</sub>X* + &delta;<sub>2</sub>Y + &delta;<sub>2+j</sub>C<sub>j</sub> <br> logit(P(S=1)) = &beta;<sub>0</sub> + &beta;<sub>1</sub>X* + &beta;<sub>2</sub>Y + &beta;<sub>2+j</sub>C<sub>j</sub> }}{\eqn{logit(P(U=1)) = \alpha_0 + \alpha_1 X + \alpha_2 Y \\ logit(P(X=1)) = \delta_0 + \delta_1 X^* + \delta_2 Y + \delta_{2+j} C_j \\ logit(P(S=1)) = \beta_0 + \beta_1 X^* + \beta_2 Y + \beta_{2+j} C_j }}
#'   }
#'   \item{Uncontrolled Confounding, Exposure Misclassification, and Selection Bias (Option 2)}{
#'     \ifelse{html}{\out{log(P(X=1,U=0)/P(X=0,U=0)) = &gamma;<sub>1,0</sub> + &gamma;<sub>1,1</sub>X* + &gamma;<sub>1,2</sub>Y + &gamma;<sub>1,2+j</sub>C<sub>j</sub> <br> log(P(X=0,U=1)/P(X=0,U=0)) = &gamma;<sub>2,0</sub> + &gamma;<sub>2,1</sub>X* + &gamma;<sub>2,2</sub>Y + &gamma;<sub>2,2+j</sub>C<sub>j</sub> <br> log(P(X=1,U=1)/P(X=0,U=0)) = &gamma;<sub>3,0</sub> + &gamma;<sub>3,1</sub>X* + &gamma;<sub>3,2</sub>Y + &gamma;<sub>3,2+j</sub>C<sub>j</sub> <br> logit(P(S=1)) = &beta;<sub>0</sub> + &beta;<sub>1</sub>X* + &beta;<sub>2</sub>Y + &beta;<sub>2+j</sub>C<sub>j</sub> }}{\eqn{log(P(X=1,U=0)/P(X=0,U=0)) = \gamma_{1,0} + \gamma_{1,1} X^* + \gamma_{1,2} Y + \gamma_{1,2+j} C_j \\ log(P(X=0,U=1)/P(X=0,U=0)) = \gamma_{2,0} + \gamma_{2,1} X^* + \gamma_{2,2} Y + \gamma_{2,2+j} C_j \\ log(P(X=1,U=1)/P(X=0,U=0)) = \gamma_{3,0} + \gamma_{3,1} X^* + \gamma_{3,2} Y + \gamma_{3,2+j} C_j \\ logit(P(S=1)) = \beta_0 + \beta_1 X^* + \beta_2 Y + \beta_{2+j} C_j }}
#'   }
#'   \item{Uncontrolled Confounding, Outcome Misclassification, and Selection Bias (Option 1)}{
#'     \ifelse{html}{\out{logit(P(U=1)) = &alpha;<sub>0</sub> + &alpha;<sub>1</sub>X + &alpha;<sub>2</sub>Y <br> logit(P(Y=1)) = &delta;<sub>0</sub> + &delta;<sub>1</sub>X + &delta;<sub>2</sub>Y* + &delta;<sub>2+j</sub>C<sub>j</sub> <br> logit(P(S=1)) = &beta;<sub>0</sub> + &beta;<sub>1</sub>X + &beta;<sub>2</sub>Y* + &beta;<sub>2+j</sub>C<sub>j</sub> }}{\eqn{logit(P(U=1)) = \alpha_0 + \alpha_1 X + \alpha_2 Y \\ logit(P(Y=1)) = \delta_0 + \delta_1 X + \delta_2 Y^* + \delta_{2+j} C_j \\ logit(P(S=1)) = \beta_0 + \beta_1 X + \beta_2 Y^* + \beta_{2+j} C_j }}
#'   }
#'   \item{Uncontrolled Confounding, Outcome Misclassification, and Selection Bias (Option 2)}{
#'     \ifelse{html}{\out{log(P(U=1,Y=0)/P(U=0,Y=0)) = &gamma;<sub>1,0</sub> + &gamma;<sub>1,1</sub>X + &gamma;<sub>1,2</sub>Y* + &gamma;<sub>1,2+j</sub>C<sub>j</sub> <br> log(P(U=0,Y=1)/P(U=0,Y=0)) = &gamma;<sub>2,0</sub> + &gamma;<sub>2,1</sub>X + &gamma;<sub>2,2</sub>Y* + &gamma;<sub>2,2+j</sub>C<sub>j</sub> <br> log(P(U=1,Y=1)/P(U=0,Y=0)) = &gamma;<sub>3,0</sub> + &gamma;<sub>3,1</sub>X + &gamma;<sub>3,2</sub>Y* + &gamma;<sub>3,2+j</sub>C<sub>j</sub> <br> logit(P(S=1)) = &beta;<sub>0</sub> + &beta;<sub>1</sub>X + &beta;<sub>2</sub>Y* + &beta;<sub>2+j</sub>C<sub>j</sub> }}{\eqn{log(P(U=1,Y=0)/P(U=0,Y=0)) = \gamma_{1,0} + \gamma_{1,1} X + \gamma_{1,2} Y^* + \gamma_{1,2+j} C_j \\ log(P(U=0,Y=1)/P(U=0,Y=0)) = \gamma_{2,0} + \gamma_{2,1} X + \gamma_{2,2} Y^* + \gamma_{2,2+j} C_j \\ log(P(U=1,Y=1)/P(U=0,Y=0)) = \gamma_{3,0} + \gamma_{3,1} X + \gamma_{3,2} Y^* + \gamma_{3,2+j} C_j \\ logit(P(S=1)) = \beta_0 + \beta_1 X + \beta_2 Y^* + \beta_{2+j} C_j }}
#'   }
#' }
#'
#' @param coef_list List of coefficient values from the above options of
#' models. Each item of the list is an equation. The left side of the equation
#' identifies the model (i.e., "u" for the model predicting the uncontrolled
#' confounder). For the multinomial models, specify the value here based on the
#' numerator (i.e., "x1u0", "x0u1", "x1u1" for the three multinomial models in
#' Uncontrolled Confounding & Exposure Misclassification, Option 2)
#' The right side of the equation is the vector of values
#' corresponding to the model coefficients (from left to right).
#'
#' @examples
#' list_for_uc <- list(
#'   u = c(-0.19, 0.61, 0.70, -0.09, 0.10, -0.15)
#' )
#'
#' bp_uc <- bias_params(coef_list = list_for_uc)
#'
#' list_for_em_om <- list(
#'   x1y0 = c(-2.18, 1.63, 0.23, 0.36),
#'   x0y1 = c(-3.17, 0.22, 1.60, 0.40),
#'   x1y1 = c(-4.76, 1.82, 1.83, 0.72)
#' )
#'
#' bp_em_om <- bias_params(coef_list = list_for_em_om)
#'
#' @export

bias_params <- function(coef_list) {
  # Check if the input is a list (R's equivalent of a dictionary)
  if (!is.list(coef_list)) {
    stop("Input must be a list (dictionary).")
  }

  # Check if all keys are strings
  if (!all(sapply(names(coef_list), is.character))) {
    stop("All keys in the dictionary must be strings.")
  }

  # Check if all values are vectors
  if (!all(sapply(coef_list, is.vector))) {
    stop("All values in the dictionary must be vectors.")
  }

  # Check if key is an acceptable value
  acceptable_values <- c(
    "u", "x", "y", "s", "x1u0", "x0u1", "x1u1", "u1y0", "u0y1", "u1y1",
    "x1y0", "x0y1", "x1y1"
  )
  for (name in names(coef_list)) {
    if (!(name %in% acceptable_values)) {
      stop(
        paste0(
          "Unacceptable value in coef_list. Acceptable values include: ",
          "u, x, y, s, x1u0, x1u1, x1u1, u1y0, u0y1, u1y1, x1y0, x0y1, x1y1"
        )
      )
    }
  }

  obj <- list(coef_list = coef_list)
  class(obj) <- "bias_params"

  return(obj)
}

#' @export

print.bias_params <- function(x, ...) {
  cat("Bias Parameters\n")
  cat("---------------------------------\n")
  for (key in names(x$coef_list)) {
    cat("Model", key, "\n")
    cat("  Coefs: ", paste(x$coef_list[[key]], collapse = ", "), "\n")
  }
  invisible(x)
}

# Adjust for exposure misclassification

# the following functions feed into adjust_em():
# adjust_em_val() (data_validation input, method: imputation),
# adjust_em_coef() (bias_params input, method: imputation)

adjust_em_val <- function(
    data_observed,
    data_validation) {
  if (!all(data_observed$confounders %in% data_validation$confounders)) {
    stop(
      "All confounders in observed data must be present in validation data.",
      call. = FALSE
    )
  }

  if (is.null(data_validation$misclassified_exposure)) {
    stop(
      paste0(
        "Attempting to adjust for a misclassified exposure.",
        "\n",
        "Validation data must have a true and misclassified exposure specified."
      ),
      call. = FALSE
    )
  }

  n <- nrow(data_observed$data)

  df <- data.frame(
    Xstar = data_observed$data[, data_observed$exposure],
    Y = data_observed$data[, data_observed$outcome]
  )
  df <- bind_cols(
    df,
    data_observed$data %>%
      select(all_of(data_observed$confounders))
  )

  if (all(df$Y %in% 0:1)) {
    y_binary <- TRUE
  } else {
    y_binary <- FALSE
  }

  df_val <- data.frame(
    X = data_validation$data[, data_validation$true_exposure],
    Y = data_validation$data[, data_validation$true_outcome],
    Xstar = data_validation$data[, data_validation$misclassified_exposure]
  )
  df_val <- bind_cols(
    df_val,
    data_validation$data %>%
      select(all_of(data_validation$confounders))
  )

  force_match(
    df$Y,
    df_val$Y,
    "Outcomes from both datasets must both be binary or both be continuous."
  )
  force_binary(
    df$Xstar,
    "Exposure in observed data must be a binary integer."
  )
  force_binary(
    df_val$Xstar,
    "Misclassified exposure in validation data must be a binary integer."
  )
  force_binary(
    df_val$X,
    "True exposure in validation data must be a binary integer."
  )

  x_mod <- glm(X ~ Xstar + Y + .,
    family = binomial(link = "logit"),
    data = df_val
  )

  x_mod_coefs <- coef(x_mod)
  x_mod_se <- sqrt(diag(vcov(x_mod)))

  # Sample coefficients independently using their standard errors
  x_mod_coefs_sampled <- rnorm(length(x_mod_coefs), x_mod_coefs, x_mod_se)
  x_pred <- x_mod_coefs_sampled[1]

  for (i in 2:length(x_mod_coefs_sampled)) {
    var_name <- names(x_mod_coefs)[i]
    x_pred <- x_pred + df[[var_name]] * x_mod_coefs_sampled[i]
  }

  df$Xpred <- rbinom(n, 1, plogis(x_pred))

  if (y_binary) {
    final <- glm(
      Y ~ Xpred + . - Xstar,
      family = binomial(link = "logit"),
      data = df
    )
  } else {
    final <- lm(
      Y ~ Xpred + . - Xstar,
      data = df
    )
  }

  return(final)
}


adjust_em_coef <- function(
    data_observed,
    x_model_coefs) {
  data <- data_observed$data
  n <- nrow(data)
  confounders <- data_observed$confounders
  len_c <- length(confounders)
  len_x_coefs <- length(x_model_coefs)

  xstar <- data[, data_observed$exposure]
  y <- data[, data_observed$outcome]

  force_binary(xstar, "Exposure must be a binary integer.")
  force_len(
    len_x_coefs,
    3 + len_c,
    paste0(
      "Incorrect length of X model coefficients. ",
      "Length should equal 3 + number of confounders."
    )
  )

  if (all(y %in% 0:1)) {
    y_binary <- TRUE
  } else {
    y_binary <- FALSE
  }

  # Extract X model coefficients
  x1_0 <- x_model_coefs[1]
  x1_xstar <- x_model_coefs[2]
  x1_y <- x_model_coefs[3]
  x1_c <- x_model_coefs[4:len_x_coefs]

  # Create base dataframe
  df <- data.frame(Xstar = xstar, Y = y)

  # Add confounders if they exist
  if (!is.null(confounders)) {
    for (i in seq_along(confounders)) {
      df[[paste0("C", i)]] <- data[, confounders[i]]
    }
  }

  # Calculate X predictions
  x_lp <- x1_0 + x1_xstar * df$Xstar + x1_y * df$Y
  if (!is.null(confounders)) {
    C_matrix <- as.matrix(
      df[, paste0("C", seq_along(confounders)), drop = FALSE]
    )
    x_lp <- x_lp + C_matrix %*% x1_c
  }
  df$Xpred <- rbinom(n, 1, plogis(x_lp))

  # Construct final model formula
  model_terms <- c("Xpred")
  if (!is.null(confounders)) {
    model_terms <- c(model_terms, paste0("C", seq_along(confounders)))
  }
  model_formula <- as.formula(
    paste("Y ~", paste(model_terms, collapse = " + "))
  )

  # Fit final model
  if (y_binary) {
    final <- glm(
      model_formula,
      family = binomial(link = "logit"),
      data = df
    )
  } else {
    final <- lm(
      model_formula,
      data = df
    )
  }

  return(final)
}


adjust_em <- function(
    data_observed,
    data_validation = NULL,
    bias_params = NULL,
    level = 0.95) {
  data <- data_observed$data
  xstar <- data[, data_observed$exposure]
  y <- data[, data_observed$outcome]

  if (all(y %in% 0:1)) {
    y_binary <- TRUE
  } else {
    y_binary <- FALSE
  }

  if (!is.null(data_validation)) {
    final <- adjust_em_val(
      data_observed,
      data_validation
    )
  } else if (!is.null(bias_params)) {
    if (is.null(bias_params$coef_list$x)) {
      stop(
        "bias_params must specify parameters for x to adjust for exposure misclassification",
        call. = FALSE
      )
    }
    final <- adjust_em_coef(
      data_observed,
      x_model_coefs = bias_params$coef_list$x
    )
  }

  calculate_results(final, level, y_binary)
}

---
title: "Multibias example: Evans"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Multibias example: Evans}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

This vignette demonstrates how to use `multibias` to adjust for uncontrolled confounding in a real-world dataset from the Evans County Heart Study. It specifically showcases how to reason and derive bias parameters that can be used to adjust for the uncontrolled confounding. We'll examine the relationship between smoking and coronary heart disease (CHD), showing how failing to account for age as a confounder can bias our estimates and how `multibias` can be used to arrive at the unbiased effect estimate despite missing data on the confounder.

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup, message = FALSE}
library(multibias)
library(dplyr)
```

The Evans County Heart Study was a prospective cohort study conducted in Evans County, Georgia, from 1960 to 1969. The study aimed to investigate risk factors for cardiovascular disease in a rural population. For this example, we'll use a subset of the data focusing on 609 participants aged 40 and older.

The key variables in our analysis are:

* `SMK`: Smoking status (1 = smoker, 0 = non-smoker)
* `CHD`: Coronary heart disease status (1 = present, 0 = absent)
* `HPT`: Hypertension status (1 = present, 0 = absent)
* `AGE`: Age in years

Let's load and examine the data:

```{r, eval = TRUE}
evans <- read.csv("evans.csv")

summary_stats <- evans %>%
  summarise(
    n = n(),
    mean_age = mean(AGE),
    sd_age = sd(AGE),
    prop_smokers = mean(SMK),
    prop_chd = mean(CHD),
    prop_hpt = mean(HPT)
  )

print(summary_stats)
```

For purposes of demonstrating `multibias`, let's pretend that our data was missing information on the confounding variable, AGE. Let's create our `data_observed` object with uncontrolled confounding (bias = "uc") and inspect the biased SMK-CHD effect estimate, adjusted for hypertension (HPT).

```{r, eval = TRUE}
df_obs <- data_observed(
  data = evans,
  bias = "uc",
  exposure = "SMK",
  outcome = "CHD",
  confounders = "HPT"
)

print(df_obs)
summary(df_obs)
```

Can we anticipate whether this odds ratio without age-adjustment is biased towards or away from the null? Let's consider the association of the uncontrolled confounder with the exposure and outcome.

```{r}
cor(evans$SMK, evans$AGE)
cor(evans$CHD, evans$AGE)
```

In our data, AGE has a negative association with SMK (older people are **less** likely to be smokers) and a positive association with CHD (older people are **more** likely to have CHD). These opposite associations must be biasing the odds ratio towards the null, creating a distortion where those who are less likely to smoke are more likely to experience the outcome.

We'll treat AGE as a binary indicator of over (1) or under (0) age 60. To adjust for the uncontrolled confounding from AGE, let's refer to the appropriate bias model for a binary uncontrolled confounder: logit(P(U=1)) = &alpha;<sub>0</sub> + &alpha;<sub>1</sub>X + &alpha;<sub>2</sub>Y + &alpha;<sub>2+j</sub>C<sub>j</sub>.

To derive the necessary bias parameters, let's make the following assumption:

* the odds of an age>60 is half as likely in smokers than non-smokers
* the odds of an age>60 is 2.5x in those with CHD than those without CHD
* the odds of an age>60 is 2x in those with HPT than those without HPT

To convert these relationships as parameters in the model, we'll log-transform them from odds ratios. For the model intercept, we can use the following reasoning: what is the probability that a non-smoker (X=0) without CHD (Y=0) and HPT (C=0) is over age 60 in this population? We'll assume this is a 25% probability. We'll use the inverse logit function `qlogis()` from the `stats` package to convert this from a probability to the intercept coefficient of the logistic regression model.

```{r, eval = TRUE}
u_0 <- qlogis(0.25)
u_x <- log(0.5)
u_y <- log(2.5)
u_c <- log(2)

u_coefs <- list(u = c(u_0, u_x, u_y, u_c))
```

Now let's plug these bias parameters into `multibias_adjust()` along with our `data_observed` object to obtain a bias-adjusted effect estimate.

```{r, eval = TRUE}
set.seed(1234)
multibias_adjust(
  data_observed = df_obs,
  bias_params = bias_params(coef_list = u_coefs),
  bootstrap = TRUE,
  bootstrap_reps = 100
)
```
We get an odds ratio of 2.2. This matches our expectation that the bias adjustment would pull the odds ratio away from the null. How does this result compare to the result we would get if age **wasn't** missing in the data and was incorporated in the outcome regression?

```{r, eval = TRUE}
full_model <- glm(CHD ~ SMK + HPT + AGE,
  family = binomial(link = "logit"),
  data = evans
)
or <- round(exp(coef(full_model)[2]), 2)
or_ci_low <- round(
  exp(coef(full_model)[2] - 1.96 * summary(full_model)$coef[2, 2]), 2
)
or_ci_high <- round(
  exp(coef(full_model)[2] + 1.96 * summary(full_model)$coef[2, 2]), 2
)

print(paste0("Odds Ratio: ", or))
print(paste0("95% CI: (", or_ci_low, ", ", or_ci_high, ")"))
```

Based on these results, it appears that the bias-adjusted odds ratio obtained via `multibias` is close to this complete-data odds ratio of 2.3.

---
title: "Frequently Asked Questions"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Frequently Asked Questions}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

# How does `multibias` fit into the causal inference workflow?

If the causal analysis process involves the following general steps:

1. Specify a causal question
2. Graph the causal assumptions (via a DAG)
3. Apply confounder balancing (matching, weighting, etc.)
4. Assess confounder balance
5. Estimate the causal effect
6. Perform sensitivity analysis of the effect estimate

Then `multibias` plugs into the final step of running sensitivity analyses.
Traditionally, the sensitivity analyses involve re-estimating results under
different assumptions related to the specified or unspecified confounders via
technique such as E-values and tipping point analyses. However, we know that
confounding is just one of several different types of biases that may impact
causal estimates, and `multibias` helps fill this void. It does so by allowing
for the simultaneous adjustment of a variety of biases: uncontrolled
confounding, exposure misclassification, outcome misclassification,
and selection bias. This allows for a more transparent and complete
understanding of how **all** potential sources of systematic error may be
impacting the analysis.

# My causal analysis leveraged propensity scores - how does that fit into a multibias analysis?

While propensity score methods are valuable for handling confounding by measured
variables, they don't address all potential sources of bias. This is where
multibias analysis becomes a crucial next step.

A limitation of propensity scores is that they can only balance covariates
that were actually measured and included in the propensity score model. A
multibias analysis allows you to quantitatively assess the potential impact
of such unmeasured confounding on top of the propensity score-adjusted
analysis. In addition, propensity scores are designed specifically for
confounding control. They do not inherently address other types of systematic
error, such as selection bias and misclassification.

In practice, you would first conduct your primary analysis using propensity
scores to control for **measured** confounders. You would then use
`multibias` to perform a sensitivity analysis for **unmeasured** confounders
and other potential sources of bias.

# What is bootstrapping and how is it relevant here?

Bootstrapping is a statistical technique used to estimate the uncertainty
around an estimate. Instead of analyzing your study data once, bootstrapping
involves repeatedly drawing random samples with replacement from your original
data. Each of these samples is the same size as the original data. For each
bootstrap sample, you then calculate the statistic of interest - a bias-adjusted
estimate in the context of `multibias`. By repeating this resampling,
you end up with a distribution of the statistic. From this distribution you can
derive the confidence interval and standard error.

When `multibias` adjusts the data's exposure-outcome effect for potential
biases, the adjusted estimate has uncertainty. This uncertainty comes not only
from the original random (sampling) error in the data, but also from the
uncertainty in the bias parameters (systematic error). **Bootstrapping is particularly important for `multibias` because it allows for confidence intervals that incorporate two sources of uncertainty: uncertainty due to random error and uncertainty due to systematic error.**

This exercise of performing bias analysis with uncertainty in the bias
parameters is called probabilistic bias analysis. When running
`multibias_adjust()` with validation data (a `data_validation` object)
across bootstrap samples, the function automatically resamples from the Normal
distribution bias parameters (estimate and standard error) inferred from the
validation data. When running `multibias_adjust()` with specified bias
parameters (a `bias_params` object) across bootstrap samples, you must input the parameter values as statistical distributions (e.g., by using the `rnorm()` or
`runif()` function).

# How do I cite `multibias`?

To get the most up-to-date citation information, please use the built-in `citation()` command in R.

```{r}
citation("multibias")
```

# Does `multibias` support time-to-event data and analyses?

Not currently

---
title: "Multibias Validation"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Multibias Validation}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup}
library(multibias)
```

This example demonstrates how `multibias` is validated using simulation data. Here we are interested in quantifying the effect of exposure *X* on outcome *Y*. The causal system can be represented in the following directed acyclic graph (DAG):

```{r out.width = '70%', echo = FALSE}
knitr::include_graphics("img/uc_emc_sel_DAG.png")
```

The variables are defined:

* X: true, unmeasured exposure
* Y: outcome
* C: measured confounder(s)
* U: unmeasured confounder
* X*: misclassified, measured exposure
* S: study selection

The DAG indicates that there are three sources of bias:
1. There is uncontrolled confounding from (unobserved) variable U.
2. The true exposure, X, is unobserved, and the misclassified exposure X* is dependent on both the exposure and outcome.
3. Lastly, there is collider stratification at variable S since exposure and outcome both affect selection. The study naturally only assesses those who were selected into the study (i.e. those with S=1), which represents a fraction of all people in the source population from which we are trying to draw inference.

A simulated dataframe corresponding to this DAG, `df_uc_emc_sel` can be loaded from the `multibias` package.

```{r, eval = TRUE}
head(df_uc_em_sel)
```

In this data, the true, unbiased exposure-outcome odds ratio (OR<sub>YX</sub>) equals ~2. However, when we run a logistic regression of the outcome on the exposure and confounders, we do not observe an odds ratio of 2 due to the multiple bias sources. We create our `data_observed` object, specifying the three biases of interest, and observe the biased model results.

```{r, eval = TRUE}
df_observed <- data_observed(
  data = df_uc_em_sel,
  bias = c("uc", "em", "sel"),
  exposure = "Xstar",
  outcome = "Y",
  confounders = c("C1", "C2", "C3")
)

print(df_observed)
summary(df_observed)
```

The function `multibias_adjust()` can be used to "reconstruct" the unbiased data and return the exposure-outcome odds ratio that would be observed in the unbiased setting.

Models for the missing variables (*U*, *X*, *S*) are used to facilitate this data reconstruction. For the above DAG, the corresponding bias models are:

* logit(P(U=1)) = &alpha;<sub>0</sub> + &alpha;<sub>1</sub>X + &alpha;<sub>2</sub>Y
* logit(P(X=1)) = &delta;<sub>0</sub> + &delta;<sub>1</sub>X* + &delta;<sub>2</sub>Y + &delta;<sub>2+j</sub>C<sub>j</sub>
* logit(P(S=1)) = &beta;<sub>0</sub> + &beta;<sub>1</sub>X* + &beta;<sub>2</sub>Y + &beta;<sub>2+j</sub>C<sub>j</sub>

where j indicates the number of measured confounders.

To perform the bias adjustment, it is necessary to obtain values of these bias parameters. Potential sources of these bias parameters may include validation data, estimates in the literature, and expert opinion. For purposes of demonstrating the methodology, we will obtain the exact values of these bias parameters. This is possible because we have access to the data of missing values that would otherwise be absent in real-world practice. This source data is available in multibias as `df_uc_emc_sel_source`.

```{r, eval = TRUE}
u_model <- glm(U ~ X + Y,
  family = binomial(link = "logit"),
  data = df_uc_em_sel_source
)
x_model <- glm(X ~ Xstar + Y + C1 + C2 + C3,
  family = binomial(link = "logit"),
  data = df_uc_em_sel_source
)
s_model <- glm(S ~ Xstar + Y + C1 + C2 + C3,
  family = binomial(link = "logit"),
  data = df_uc_em_sel_source
)
```

In this example we'll perform probabilistic bias analysis, representing each bias parameter as a draw from a Normal distribution. This allows us to account for our uncertainty in the "true" value of the bias parameters. We will leverage bootstrap sampling, which is built into `multibias_adjust()` to obtain a confidence interval that incorporates both uncertainty from random error and uncertainty from systematic error (our bias parameters). Each loop iteration will thus have slightly different values for the bias parameters.

We create our `bias_params` object, bringing in the model results we determined above.

```{r}
bp <- bias_params(
  coef_list = list(
    u = c(
      rnorm(1, mean = u_model$coef[1], sd = summary(u_model)$coef[1, 2]),
      rnorm(1, mean = u_model$coef[2], sd = summary(u_model)$coef[2, 2]),
      rnorm(1, mean = u_model$coef[3], sd = summary(u_model)$coef[3, 2])
    ),
    x = c(
      rnorm(1, mean = x_model$coef[1], sd = summary(x_model)$coef[1, 2]),
      rnorm(1, mean = x_model$coef[2], sd = summary(x_model)$coef[2, 2]),
      rnorm(1, mean = x_model$coef[3], sd = summary(x_model)$coef[3, 2]),
      rnorm(1, mean = x_model$coef[4], sd = summary(x_model)$coef[4, 2]),
      rnorm(1, mean = x_model$coef[5], sd = summary(x_model)$coef[5, 2]),
      rnorm(1, mean = x_model$coef[6], sd = summary(x_model)$coef[6, 2])
    ),
    s = c(
      rnorm(1, mean = s_model$coef[1], sd = summary(s_model)$coef[1, 2]),
      rnorm(1, mean = s_model$coef[2], sd = summary(s_model)$coef[2, 2]),
      rnorm(1, mean = s_model$coef[3], sd = summary(s_model)$coef[3, 2]),
      rnorm(1, mean = s_model$coef[4], sd = summary(s_model)$coef[4, 2]),
      rnorm(1, mean = s_model$coef[5], sd = summary(s_model)$coef[5, 2]),
      rnorm(1, mean = s_model$coef[6], sd = summary(s_model)$coef[6, 2])
    )
  )
)
```

The `data_observed` and `bias_params` objects are specified as inputs of `multibias`. In addition, we will indicate that we want results ran over bootstrap samples.

```{r}
multibias_adjust(
  data_observed = df_observed,
  bias_params = bp,
  bootstrap = TRUE,
  bootstrap_reps = 10,
  level = 0.95
)
```

We arrive at an estimate of the effect of X on Y of approximately 2, correctly corresponding to the relationship that was built into the data. This showcases the power of multibias to allow researchers to arrive at valid causal effect estimates in the presence of missing or incomplete data.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data_em_sel.R
\docType{data}
\name{df_em_sel_source}
\alias{df_em_sel_source}
\title{Data source for \code{df_em_sel}}
\format{
A dataframe with 100,000 rows and 7 columns:
\describe{
\item{X}{true exposure, 1 = present and 0 = absent}
\item{Y}{outcome, 1 = present and 0 = absent}
\item{C1}{1st confounder, 1 = present and 0 = absent}
\item{C2}{2nd confounder, 1 = present and 0 = absent}
\item{C3}{3rd confounder, 1 = present and 0 = absent}
\item{Xstar}{misclassified exposure, 1 = present and 0 = absent}
\item{S}{selection, 1 = selected into the study and 0 = not selected into the study}
}
}
\usage{
df_em_sel_source
}
\description{
Data with complete information on the two sources of bias, three known
confounders, and 100,000 observations. This data is used to derive
\code{df_em_sel} and can be used to obtain bias parameters for purposes
of validating the simultaneous multi-bias adjustment method with
\code{df_em_sel}. With this source data, the fitted regression
\ifelse{html}{\out{logit(P(Y=1)) = &alpha;<sub>0</sub> + &alpha;<sub>1</sub>X + &alpha;<sub>2</sub>C1 + &alpha;<sub>3</sub>C2 + &alpha;<sub>4</sub>C3}}{\eqn{logit(P(Y=1)) = \alpha_0 + \alpha_1 X + \alpha_2 C1 + \alpha_3 C2 + \alpha_4 C3}}
shows that the true, unbiased exposure-outcome odds ratio = 2.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data_uc.R
\docType{data}
\name{df_uc_source}
\alias{df_uc_source}
\title{Data source for \code{df_uc}}
\format{
A dataframe with 100,000 rows and 8 columns:
\describe{
\item{X_bi}{binary exposure, 1 = present and 0 = absent}
\item{X_cont}{continuous exposure}
\item{Y_bi}{binary outcome corresponding to exposure \emph{X_bi}, 1 = present and 0 = absent}
\item{Y_cont}{continuous outcome corresponding to exposure \emph{X_cont}}
\item{C1}{1st confounder, 1 = present and 0 = absent}
\item{C2}{2nd confounder, 1 = present and 0 = absent}
\item{C3}{3rd confounder, 1 = present and 0 = absent}
\item{U}{uncontrolled confounder, 1 = present and 0 = absent}
}
}
\usage{
df_uc_source
}
\description{
Data with complete information on one source of bias, three known
confounders, and 100,000 observations. This data is used to derive
\code{df_uc} and can be used to obtain bias parameters for purposes
of validating the simultaneous multi-bias adjustment method with
\code{df_uc}. With this source data, the fitted regression
\ifelse{html}{\out{g(P(Y=1)) = &alpha;<sub>0</sub> + &alpha;<sub>1</sub>X + &alpha;<sub>2</sub>C1 + &alpha;<sub>3</sub>C2 + &alpha;<sub>4</sub>C3 + &alpha;<sub>5</sub>U}}{\eqn{logit(P(Y=1)) = \alpha_0 + \alpha_1 X + \alpha_2 C1 + \alpha_3 C2 + \alpha_4 C3 + \alpha_5 U}}
shows that the true, unbiased exposure-outcome effect estimate = 2 when:
\enumerate{
\item g = logit, Y = \emph{Y_bi}, and X = \emph{X_bi} or
\item g = identity, Y = \emph{Y_cont}, X = \emph{X_cont}.
}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/causal_data.R
\name{data_observed}
\alias{data_observed}
\title{Represent observed causal data}
\usage{
data_observed(data, bias, exposure, outcome, confounders = NULL)
}
\arguments{
\item{data}{Dataframe for bias analysis.}

\item{bias}{String type(s) of bias distorting the effect of the exposure
on the outcome. Can choose from a subset of the following: "uc", "em", "om",
"sel". These correspond to uncontrolled confounding, exposure
misclassification, outcome misclassification, and selection bias,
respectively.}

\item{exposure}{String name of the column in \code{data} corresponding to the
exposure variable.}

\item{outcome}{String name of the column in \code{data} corresponding to the
outcome variable.}

\item{confounders}{String name(s) of the column(s) in \code{data} corresponding
to the confounding variable(s).}
}
\value{
An object of class \code{data_observed} containing:
\item{data}{A dataframe with the selected columns}
\item{bias}{The type(s) of bias present}
\item{exposure}{The name of the exposure variable}
\item{outcome}{The name of the outcome variable}
\item{confounders}{The name(s) of the confounder variable(s)}
}
\description{
\code{data_observed} combines the observed dataframe with specific identification
of the columns corresponding to the exposure, outcome, and confounders. It is
an essential input of the \code{\link[=multibias_adjust]{multibias_adjust()}} function.
}
\examples{
df <- data_observed(
  data = df_uc,
  bias = "uc",
  exposure = "X_bi",
  outcome = "Y_bi",
  confounders = c("C1", "C2", "C3")
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/multibias-package.R
\docType{package}
\name{multibias-package}
\alias{multibias}
\alias{multibias-package}
\title{multibias: Multiple Bias Analysis in Causal Inference}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

Quantify exposure-outcome causal effects with adjustment for multiple biases. The functions can simultaneously adjust for any combination of uncontrolled confounding, exposure/outcome misclassification, and selection bias. The underlying method generalizes the combination of inverse probability of selection weighting with predictive value weighting. Simultaneous multi-bias analysis can be used to enhance the validity and transparency of real-world evidence obtained from observational, longitudinal studies. Based on the work from Paul Brendel, Aracelis Torres, and Onyebuchi Arah (2023) \doi{10.1093/ije/dyad001}.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/pcbrendel/multibias}
  \item \url{http://www.paulbrendel.com/multibias/}
  \item Report bugs at \url{https://github.com/pcbrendel/multibias/issues}
}

}
\author{
\strong{Maintainer}: Paul Brendel \email{pcbrendel@gmail.com} [copyright holder]

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data_em.R
\docType{data}
\name{df_em}
\alias{df_em}
\title{Simulated data with exposure misclassification}
\format{
A dataframe with 100,000 rows and 5 columns:
\describe{
\item{Xstar}{misclassified exposure, 1 = present and 0 = absent}
\item{Y}{outcome, 1 = present and 0 = absent}
\item{C1}{1st confounder, 1 = present and 0 = absent}
\item{C2}{2nd confounder, 1 = present and 0 = absent}
\item{C3}{3rd confounder, 1 = present and 0 = absent}
}
}
\usage{
df_em
}
\description{
Data containing one source of bias, three known confounders, and
100,000 observations. This data is obtained from \code{df_emc_source}
by removing the column \emph{X}. The resulting data corresponds to
what a researcher would see in the real-world: a misclassified exposure,
\emph{Xstar}, and no data on the true exposure. As seen in
\code{df_emc_source}, the true, unbiased exposure-outcome odds ratio = 2.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data_em_sel.R
\docType{data}
\name{df_em_sel}
\alias{df_em_sel}
\title{Simulated data with exposure misclassification and selection bias}
\format{
A dataframe with 100,000 rows and 5 columns:
\describe{
\item{Xstar}{misclassified exposure, 1 = present and 0 = absent}
\item{Y}{outcome, 1 = present and 0 = absent}
\item{C1}{1st confounder, 1 = present and 0 = absent}
\item{C2}{2nd confounder, 1 = present and 0 = absent}
\item{C3}{3rd confounder, 1 = present and 0 = absent}
}
}
\usage{
df_em_sel
}
\description{
Data containing two sources of bias, three known confounders, and
100,000 observations. This data is obtained by sampling with replacement
with probability = \emph{S} from \code{df_em_sel_source} then removing the
columns \emph{X} and \emph{S}. The resulting data corresponds to what a
researcher would see in the real-world: a misclassified exposure,
\emph{Xstar}, and missing data for those not selected into the study
(\emph{S}=0). As seen in \code{df_em_sel_source}, the true, unbiased
exposure-outcome odds ratio = 2.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data_uc_em.R
\docType{data}
\name{df_uc_em}
\alias{df_uc_em}
\title{Simulated data with uncontrolled confounding and exposure misclassification}
\format{
A dataframe with 100,000 rows and 5 columns:
\describe{
\item{Xstar}{misclassified exposure, 1 = present and 0 = absent}
\item{Y}{outcome, 1 = present and 0 = absent}
\item{C1}{1st confounder, 1 = present and 0 = absent}
\item{C2}{2nd confounder, 1 = present and 0 = absent}
\item{C3}{3rd confounder, 1 = present and 0 = absent}
}
}
\usage{
df_uc_em
}
\description{
Data containing two sources of bias, three known confounders, and
100,000 observations. This data is obtained from \code{df_uc_em_source}
by removing the columns \emph{X} and \emph{U}. The resulting data
corresponds to what a researcher would see in the real-world: a
misclassified exposure, \emph{Xstar}, and missing data on a confounder
\emph{U}. As seen in \code{df_uc_em_source}, the true, unbiased
exposure-outcome odds ratio = 2.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/multibias_plot.R
\name{multibias_plot}
\alias{multibias_plot}
\title{Create a Forest Plot comparing observed and adjusted effect estimates}
\usage{
multibias_plot(data_observed, multibias_result_list, log_scale = FALSE)
}
\arguments{
\item{data_observed}{Object of class \code{data_observed} representing the
observed causal data and effect of interest.}

\item{multibias_result_list}{A named list of sensitivity analysis results.
Each element should be a result from \code{\link[=multibias_adjust]{multibias_adjust()}}.}

\item{log_scale}{Boolean indicating whether to display the x-axis on the
log scale. Default is FALSE.}
}
\value{
A ggplot object showing a forest plot with:
\itemize{
\item Point estimates (blue dots)
\item Confidence intervals (gray horizontal lines)
\item A vertical reference line at x=1 (dashed)
\item Appropriate labels and title
}
}
\description{
This function generates a forest plot comparing the observed effect estimate
with adjusted effect estimates from sensitivity analyses. The plot includes
point estimates and confidence intervals for each analysis.
}
\examples{
\dontrun{
df_observed <- data_observed(
  data = df_em,
  bias = "em",
  exposure = "Xstar",
  outcome = "Y",
  confounders = "C1"
)

bp1 <- bias_params(coef_list = list(x = c(-2.10, 1.62, 0.63, 0.35)))
bp2 <- bias_params(coef_list = list(x = c(-2.10 * 2, 1.62 * 2, 0.63 * 2, 0.35 * 2)))

result1 <- multibias_adjust(
  data_observed = df_observed,
  bias_params = bp1
)
result2 <- multibias_adjust(
  data_observed = df_observed,
  bias_params = bp2
)

multibias_plot(
  data_observed = df_observed,
  multibias_result_list = list(
    "Adjusted with bias params" = result1,
    "Adjusted with bias params doubled" = result2
  )
)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data_uc_sel.R
\docType{data}
\name{df_uc_sel}
\alias{df_uc_sel}
\title{Simulated data with uncontrolled confounding and selection bias}
\format{
A dataframe with 100,000 rows and 5 columns:
\describe{
\item{X}{exposure, 1 = present and 0 = absent}
\item{Y}{outcome, 1 = present and 0 = absent}
\item{C1}{1st confounder, 1 = present and 0 = absent}
\item{C2}{2nd confounder, 1 = present and 0 = absent}
\item{C3}{3rd confounder, 1 = present and 0 = absent}
}
}
\usage{
df_uc_sel
}
\description{
Data containing two sources of bias, three known confounders, and 100,000
observations. This data is obtained by sampling with replacement with
probability = \emph{S} from \code{df_uc_sel_source} then removing
the columns \emph{U} and \emph{S}. The resulting data corresponds to
what a researcher would see
in the real-world: missing data on confounder \emph{U}; and missing data for
those not selected into the study (\emph{S}=0). As seen in
\code{df_uc_sel_source}, the true, unbiased exposure-outcome odds ratio = 2.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/causal_data.R
\name{summary.data_observed}
\alias{summary.data_observed}
\title{Summary method for data_observed objects}
\usage{
\method{summary}{data_observed}(object, ...)
}
\arguments{
\item{object}{A \code{data_observed} object}

\item{...}{Additional arguments passed to summary}
}
\value{
A data frame containing model coefficients, standard errors,
confidence intervals, and p-values. For binary outcomes,
coefficients are exponentiated to show odds ratios.
}
\description{
Provides a statistical summary of the observed data by fitting either:
\itemize{
\item A logistic regression model for binary outcomes
\item A linear regression model for continuous outcomes
}

The model includes the exposure and all confounders as predictors.
For binary outcomes, estimates are exponentiated to show odds ratios.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/causal_data.R
\name{print.data_observed}
\alias{print.data_observed}
\title{Print method for data_observed objects}
\usage{
\method{print}{data_observed}(x, ...)
}
\arguments{
\item{x}{A \code{data_observed} object}

\item{...}{Additional arguments passed to print}
}
\value{
The input object invisibly, allowing for method chaining
}
\description{
Prints a formatted summary of a \code{data_observed} object, including:
\itemize{
\item The types of biases present
\item The exposure, outcome, and confounder variables
\item A preview of the first 5 rows of data
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bias_params.R
\name{bias_params}
\alias{bias_params}
\title{Represent bias parameters}
\usage{
bias_params(coef_list)
}
\arguments{
\item{coef_list}{List of coefficient values from the above options of
models. Each item of the list is an equation. The left side of the equation
identifies the model (i.e., "u" for the model predicting the uncontrolled
confounder). For the multinomial models, specify the value here based on the
numerator (i.e., "x1u0", "x0u1", "x1u1" for the three multinomial models in
Uncontrolled Confounding & Exposure Misclassification, Option 2)
The right side of the equation is the vector of values
corresponding to the model coefficients (from left to right).}
}
\description{
\code{bias_params} is one of two different options to represent bias assumptions for bias adjustment. The \code{\link[=multibias_adjust]{multibias_adjust()}} function will apply the assumptions from these models and use them to adjust for biases in the observed data. It takes one input, a list, where each item in the list corresponds to the necessary models for bias adjustment. See below for bias models.

For each of the following bias models, the variables are defined:
\itemize{
\item X = True exposure
\item X* = Misclassified exposure
\item Y = True outcome
\item Y* = Misclassified outcome
\item C = Known confounder
\item j = Number of known confounders
\item U = Uncontrolled confounder
\item S = Selection indicator
}

\describe{
\item{Uncontrolled confounding}{
\ifelse{html}{\out{logit(P(U=1)) = &alpha;<sub>0</sub> + &alpha;<sub>1</sub>X + &alpha;<sub>2</sub>Y + &alpha;<sub>2+j</sub>C<sub>j</sub> }}{\eqn{logit(P(U=1)) = \alpha_0 + \alpha_1 X + \alpha_2 Y + \alpha_{2+j} C_j }}
}

\item{Exposure misclassification}{
\ifelse{html}{\out{logit(P(X=1)) = &delta;<sub>0</sub> + &delta;<sub>1</sub>X* + &delta;<sub>2</sub>Y + &delta;<sub>2+j</sub>C<sub>j</sub> }}{\eqn{logit(P(X=1)) = \delta_0 + \delta_1 X^* + \delta_2 Y + \delta_{2+j} C_j }}
}

\item{Outcome misclassification}{
\ifelse{html}{\out{logit(P(Y=1)) = &delta;<sub>0</sub> + &delta;<sub>1</sub>X + &delta;<sub>2</sub>Y* + &delta;<sub>2+j</sub>C<sub>j</sub> }}{\eqn{logit(P(Y=1)) = \delta_0 + \delta_1 X + \delta_2 Y^* + \delta_{2+j} C_j }}
}

\item{Selection bias}{
\ifelse{html}{\out{logit(P(S=1)) = &beta;<sub>0</sub> + &beta;<sub>1</sub>X + &beta;<sub>2</sub>Y }}{\eqn{logit(P(S=1)) = \beta_0 + \beta_1 X + \beta_2 Y }}
}

\item{Uncontrolled Confounding & Exposure Misclassification (Option 1)}{
\ifelse{html}{\out{logit(P(U=1)) = &alpha;<sub>0</sub> + &alpha;<sub>1</sub>X + &alpha;<sub>2</sub>Y <br> logit(P(X=1)) = &delta;<sub>0</sub> + &delta;<sub>1</sub>X* + &delta;<sub>2</sub>Y + &delta;<sub>2+j</sub>C<sub>j</sub> }}{\eqn{logit(P(U=1)) = \alpha_0 + \alpha_1 X + \alpha_2 Y \\ logit(P(X=1)) = \delta_0 + \delta_1 X^* + \delta_2 Y + \delta_{2+j} C_j }}
}

\item{Uncontrolled Confounding & Exposure Misclassification (Option 2)}{
\ifelse{html}{\out{log(P(X=1,U=0)/P(X=0,U=0)) = &gamma;<sub>1,0</sub> + &gamma;<sub>1,1</sub>X* + &gamma;<sub>1,2</sub>Y + &gamma;<sub>1,2+j</sub>C<sub>j</sub> <br> log(P(X=0,U=1)/P(X=0,U=0)) = &gamma;<sub>2,0</sub> + &gamma;<sub>2,1</sub>X* + &gamma;<sub>2,2</sub>Y + &gamma;<sub>2,2+j</sub>C<sub>j</sub> <br> log(P(X=1,U=1)/P(X=0,U=0)) = &gamma;<sub>3,0</sub> + &gamma;<sub>3,1</sub>X* + &gamma;<sub>3,2</sub>Y + &gamma;<sub>3,2+j</sub>C<sub>j</sub> }}{\eqn{log(P(X=1,U=0)/P(X=0,U=0)) = \gamma_{1,0} + \gamma_{1,1} X^* + \gamma_{1,2} Y + \gamma_{1,2+j} C_j \\ log(P(X=0,U=1)/P(X=0,U=0)) = \gamma_{2,0} + \gamma_{2,1} X^* + \gamma_{2,2} Y + \gamma_{2,2+j} C_j \\ log(P(X=1,U=1)/P(X=0,U=0)) = \gamma_{3,0} + \gamma_{3,1} X^* + \gamma_{3,2} Y + \gamma_{3,2+j} C_j }}
}

\item{Uncontrolled Confounding & Outcome Misclassification (Option 1)}{
\ifelse{html}{\out{logit(P(U=1)) = &alpha;<sub>0</sub> + &alpha;<sub>1</sub>X + &alpha;<sub>2</sub>Y <br> logit(P(Y=1)) = &delta;<sub>0</sub> + &delta;<sub>1</sub>X + &delta;<sub>2</sub>Y* + &delta;<sub>2+j</sub>C<sub>j</sub> }}{\eqn{logit(P(U=1)) = \alpha_0 + \alpha_1 X + \alpha_2 Y \\ logit(P(Y=1)) = \delta_0 + \delta_1 X + \delta_2 Y^* + \delta_{2+j} C_j }}
}

\item{Uncontrolled Confounding & Outcome Misclassification (Option 2)}{
\ifelse{html}{\out{log(P(U=1,Y=0)/P(U=0,Y=0)) = &gamma;<sub>1,0</sub> + &gamma;<sub>1,1</sub>X + &gamma;<sub>1,2</sub>Y* + &gamma;<sub>1,2+j</sub>C<sub>j</sub> <br> log(P(U=0,Y=1)/P(U=0,Y=0)) = &gamma;<sub>2,0</sub> + &gamma;<sub>2,1</sub>X + &gamma;<sub>2,2</sub>Y* + &gamma;<sub>2,2+j</sub>C<sub>j</sub> <br> log(P(U=1,Y=1)/P(U=0,Y=0)) = &gamma;<sub>3,0</sub> + &gamma;<sub>3,1</sub>X + &gamma;<sub>3,2</sub>Y* + &gamma;<sub>3,2+j</sub>C<sub>j</sub> }}{\eqn{log(P(U=1,Y=0)/P(U=0,Y=0)) = \gamma_{1,0} + \gamma_{1,1} X + \gamma_{1,2} Y^* + \gamma_{1,2+j} C_j \\ log(P(U=0,Y=1)/P(U=0,Y=0)) = \gamma_{2,0} + \gamma_{2,1} X + \gamma_{2,2} Y^* + \gamma_{2,2+j} C_j \\ log(P(U=1,Y=1)/P(U=0,Y=0)) = \gamma_{3,0} + \gamma_{3,1} X + \gamma_{3,2} Y^* + \gamma_{3,2+j} C_j }}
}

\item{Uncontrolled Confounding & Selection Bias}{
\ifelse{html}{\out{logit(P(U=1)) = &alpha;<sub>0</sub> + &alpha;<sub>1</sub>X + &alpha;<sub>2</sub>Y + &alpha;<sub>2+j</sub>C<sub>j</sub> <br> logit(P(S=1)) = &beta;<sub>0</sub> + &beta;<sub>1</sub>X + &beta;<sub>2</sub>Y }}{\eqn{logit(P(U=1)) = \alpha_0 + \alpha_1 X + \alpha_2 Y + \alpha_{2+j} C_j \\ logit(P(S=1)) = \beta_0 + \beta_1 X + \beta_2 Y }}
}

\item{Exposure Misclassification & Outcome Misclassification (Option 1)}{
\ifelse{html}{\out{logit(P(X=1)) = &delta;<sub>0</sub> + &delta;<sub>1</sub>X* + &delta;<sub>2</sub>Y* + &delta;<sub>2+j</sub>C<sub>j</sub> <br> logit(P(Y=1)) = &beta;<sub>0</sub> + &beta;<sub>1</sub>X + &beta;<sub>2</sub>Y* + &beta;<sub>2+j</sub>C<sub>j</sub> }}{\eqn{logit(P(X=1)) = \delta_0 + \delta_1 X^* + \delta_2 Y^* + \delta_{2+j} C_j \\ logit(P(Y=1)) = \beta_0 + \beta_1 X + \beta_2 Y^* + \beta_{2+j} C_j }}
}

\item{Exposure Misclassification & Outcome Misclassification (Option 2)}{
\ifelse{html}{\out{log(P(X=1,Y=0) / P(X=0,Y=0)) = &gamma;<sub>1,0</sub> + &gamma;<sub>1,1</sub>X* + &gamma;<sub>1,2</sub>Y* + &gamma;<sub>1,2+j</sub>C<sub>j</sub> <br> log(P(X=0,Y=1) / P(X=0,Y=0)) = &gamma;<sub>2,0</sub> + &gamma;<sub>2,1</sub>X* + &gamma;<sub>2,2</sub>Y* + &gamma;<sub>2,2+j</sub>C<sub>j</sub> <br> log(P(X=1,Y=1) / P(X=0,Y=0)) = &gamma;<sub>3,0</sub> + &gamma;<sub>3,1</sub>X* + &gamma;<sub>3,2</sub>Y* + &gamma;<sub>3,2+j</sub>C<sub>j</sub> }}{\eqn{log(P(X=1,Y=0) / P(X=0,Y=0)) = \gamma_{1,0} + \gamma_{1,1} X^* + \gamma_{1,2} Y^* + \gamma_{1,2+j} C_j \\ log(P(X=0,Y=1) / P(X=0,Y=0)) = \gamma_{2,0} + \gamma_{2,1} X^* + \gamma_{2,2} Y^* + \gamma_{2,2+j} C_j \\ log(P(X=1,Y=1) / P(X=0,Y=0)) = \gamma_{3,0} + \gamma_{3,1} X^* + \gamma_{3,2} Y^* + \gamma_{3,2+j} C_j }}
}

\item{Exposure Misclassification & Selection Bias}{
\ifelse{html}{\out{logit(P(X=1)) = &delta;<sub>0</sub> + &delta;<sub>1</sub>X* + &delta;<sub>2</sub>Y + &delta;<sub>2+j</sub>C<sub>j</sub> <br> logit(P(S=1)) = &beta;<sub>0</sub> + &beta;<sub>1</sub>X* + &beta;<sub>2</sub>Y + &beta;<sub>2+j</sub>C<sub>j</sub> }}{\eqn{logit(P(X=1)) = \delta_0 + \delta_1 X^* + \delta_2 Y + \delta_{2+j} C_j \\ logit(P(S=1)) = \beta_0 + \beta_1 X^* + \beta_2 Y + \beta_{2+j} C_j }}
}

\item{Outcome Misclassification & Selection Bias}{
\ifelse{html}{\out{logit(P(Y=1)) = &delta;<sub>0</sub> + &delta;<sub>1</sub>X + &delta;<sub>2</sub>Y* + &delta;<sub>2+j</sub>C<sub>j</sub> <br> logit(P(S=1)) = &beta;<sub>0</sub> + &beta;<sub>1</sub>X + &beta;<sub>2</sub>Y* + &beta;<sub>2+j</sub>C<sub>j</sub> }}{\eqn{logit(P(Y=1)) = \delta_0 + \delta_1 X + \delta_2 Y^* + \delta_{2+j} C_j \\ logit(P(S=1)) = \beta_0 + \beta_1 X + \beta_2 Y^* + \beta_{2+j} C_j }}
}
\item{Uncontrolled Confounding, Exposure Misclassification, and Selection Bias (Option 1)}{
\ifelse{html}{\out{logit(P(U=1)) = &alpha;<sub>0</sub> + &alpha;<sub>1</sub>X + &alpha;<sub>2</sub>Y <br> logit(P(X=1)) = &delta;<sub>0</sub> + &delta;<sub>1</sub>X* + &delta;<sub>2</sub>Y + &delta;<sub>2+j</sub>C<sub>j</sub> <br> logit(P(S=1)) = &beta;<sub>0</sub> + &beta;<sub>1</sub>X* + &beta;<sub>2</sub>Y + &beta;<sub>2+j</sub>C<sub>j</sub> }}{\eqn{logit(P(U=1)) = \alpha_0 + \alpha_1 X + \alpha_2 Y \\ logit(P(X=1)) = \delta_0 + \delta_1 X^* + \delta_2 Y + \delta_{2+j} C_j \\ logit(P(S=1)) = \beta_0 + \beta_1 X^* + \beta_2 Y + \beta_{2+j} C_j }}
}
\item{Uncontrolled Confounding, Exposure Misclassification, and Selection Bias (Option 2)}{
\ifelse{html}{\out{log(P(X=1,U=0)/P(X=0,U=0)) = &gamma;<sub>1,0</sub> + &gamma;<sub>1,1</sub>X* + &gamma;<sub>1,2</sub>Y + &gamma;<sub>1,2+j</sub>C<sub>j</sub> <br> log(P(X=0,U=1)/P(X=0,U=0)) = &gamma;<sub>2,0</sub> + &gamma;<sub>2,1</sub>X* + &gamma;<sub>2,2</sub>Y + &gamma;<sub>2,2+j</sub>C<sub>j</sub> <br> log(P(X=1,U=1)/P(X=0,U=0)) = &gamma;<sub>3,0</sub> + &gamma;<sub>3,1</sub>X* + &gamma;<sub>3,2</sub>Y + &gamma;<sub>3,2+j</sub>C<sub>j</sub> <br> logit(P(S=1)) = &beta;<sub>0</sub> + &beta;<sub>1</sub>X* + &beta;<sub>2</sub>Y + &beta;<sub>2+j</sub>C<sub>j</sub> }}{\eqn{log(P(X=1,U=0)/P(X=0,U=0)) = \gamma_{1,0} + \gamma_{1,1} X^* + \gamma_{1,2} Y + \gamma_{1,2+j} C_j \\ log(P(X=0,U=1)/P(X=0,U=0)) = \gamma_{2,0} + \gamma_{2,1} X^* + \gamma_{2,2} Y + \gamma_{2,2+j} C_j \\ log(P(X=1,U=1)/P(X=0,U=0)) = \gamma_{3,0} + \gamma_{3,1} X^* + \gamma_{3,2} Y + \gamma_{3,2+j} C_j \\ logit(P(S=1)) = \beta_0 + \beta_1 X^* + \beta_2 Y + \beta_{2+j} C_j }}
}
\item{Uncontrolled Confounding, Outcome Misclassification, and Selection Bias (Option 1)}{
\ifelse{html}{\out{logit(P(U=1)) = &alpha;<sub>0</sub> + &alpha;<sub>1</sub>X + &alpha;<sub>2</sub>Y <br> logit(P(Y=1)) = &delta;<sub>0</sub> + &delta;<sub>1</sub>X + &delta;<sub>2</sub>Y* + &delta;<sub>2+j</sub>C<sub>j</sub> <br> logit(P(S=1)) = &beta;<sub>0</sub> + &beta;<sub>1</sub>X + &beta;<sub>2</sub>Y* + &beta;<sub>2+j</sub>C<sub>j</sub> }}{\eqn{logit(P(U=1)) = \alpha_0 + \alpha_1 X + \alpha_2 Y \\ logit(P(Y=1)) = \delta_0 + \delta_1 X + \delta_2 Y^* + \delta_{2+j} C_j \\ logit(P(S=1)) = \beta_0 + \beta_1 X + \beta_2 Y^* + \beta_{2+j} C_j }}
}
\item{Uncontrolled Confounding, Outcome Misclassification, and Selection Bias (Option 2)}{
\ifelse{html}{\out{log(P(U=1,Y=0)/P(U=0,Y=0)) = &gamma;<sub>1,0</sub> + &gamma;<sub>1,1</sub>X + &gamma;<sub>1,2</sub>Y* + &gamma;<sub>1,2+j</sub>C<sub>j</sub> <br> log(P(U=0,Y=1)/P(U=0,Y=0)) = &gamma;<sub>2,0</sub> + &gamma;<sub>2,1</sub>X + &gamma;<sub>2,2</sub>Y* + &gamma;<sub>2,2+j</sub>C<sub>j</sub> <br> log(P(U=1,Y=1)/P(U=0,Y=0)) = &gamma;<sub>3,0</sub> + &gamma;<sub>3,1</sub>X + &gamma;<sub>3,2</sub>Y* + &gamma;<sub>3,2+j</sub>C<sub>j</sub> <br> logit(P(S=1)) = &beta;<sub>0</sub> + &beta;<sub>1</sub>X + &beta;<sub>2</sub>Y* + &beta;<sub>2+j</sub>C<sub>j</sub> }}{\eqn{log(P(U=1,Y=0)/P(U=0,Y=0)) = \gamma_{1,0} + \gamma_{1,1} X + \gamma_{1,2} Y^* + \gamma_{1,2+j} C_j \\ log(P(U=0,Y=1)/P(U=0,Y=0)) = \gamma_{2,0} + \gamma_{2,1} X + \gamma_{2,2} Y^* + \gamma_{2,2+j} C_j \\ log(P(U=1,Y=1)/P(U=0,Y=0)) = \gamma_{3,0} + \gamma_{3,1} X + \gamma_{3,2} Y^* + \gamma_{3,2+j} C_j \\ logit(P(S=1)) = \beta_0 + \beta_1 X + \beta_2 Y^* + \beta_{2+j} C_j }}
}
}
}
\examples{
list_for_uc <- list(
  u = c(-0.19, 0.61, 0.70, -0.09, 0.10, -0.15)
)

bp_uc <- bias_params(coef_list = list_for_uc)

list_for_em_om <- list(
  x1y0 = c(-2.18, 1.63, 0.23, 0.36),
  x0y1 = c(-3.17, 0.22, 1.60, 0.40),
  x1y1 = c(-4.76, 1.82, 1.83, 0.72)
)

bp_em_om <- bias_params(coef_list = list_for_em_om)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data_uc.R
\docType{data}
\name{df_uc}
\alias{df_uc}
\title{Simulated data with uncontrolled confounding}
\format{
A dataframe with 100,000 rows and 7 columns:
\describe{
\item{X_bi}{binary exposure, 1 = present and 0 = absent}
\item{X_cont}{continuous exposure}
\item{Y_bi}{binary outcome corresponding to exposure \emph{X_bi}, 1 = present and 0 = absent}
\item{Y_cont}{continuous outcome corresponding to exposure \emph{X_cont}}
\item{C1}{1st confounder, 1 = present and 0 = absent}
\item{C2}{2nd confounder, 1 = present and 0 = absent}
\item{C3}{3rd confounder, 1 = present and 0 = absent}
}
}
\usage{
df_uc
}
\description{
Data containing one source of bias, three known confounders, and
100,000 observations. This data is obtained from \code{df_uc_source}
by removing the column \emph{U}. The resulting data corresponds to
what a researcher would see in the real-world: information on known
confounders (\emph{C1}, \emph{C2}, and \emph{C3}), but not for
confounder \emph{U}.
As seen in \code{df_uc_source}, the true, unbiased exposure-outcome
effect estimate = 2.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/causal_data.R
\name{data_validation}
\alias{data_validation}
\title{Represent validation causal data}
\usage{
data_validation(
  data,
  true_exposure,
  true_outcome,
  confounders = NULL,
  misclassified_exposure = NULL,
  misclassified_outcome = NULL,
  selection = NULL
)
}
\arguments{
\item{data}{Dataframe of validation data}

\item{true_exposure}{String name of the column in \code{data} corresponding to
the true exposure.}

\item{true_outcome}{String name of the column in \code{data} corresponding to
the true outcome.}

\item{confounders}{String name(s) of the column(s) in \code{data} corresponding
to the confounding variable(s).}

\item{misclassified_exposure}{String name of the column in \code{data}
corresponding to the misclassified exposure.}

\item{misclassified_outcome}{String name of the column in \code{data}
corresponding to the misclassified outcome.}

\item{selection}{String name of the column in \code{data} corresponding to the
selection indicator.}
}
\value{
An object of class \code{data_validation} containing:
\item{data}{A dataframe with the selected columns}
\item{true_exposure}{The name of the true exposure variable}
\item{true_outcome}{The name of the true outcome variable}
\item{confounders}{The name(s) of the confounder variable(s)}
\item{misclassified_exposure}{The name of the misclassified exposure variable}
\item{misclassified_outcome}{The name of the misclassified outcome variable}
\item{selection}{The name of the selection indicator variable}
}
\description{
\code{data_validation} is one of two different options to represent bias
assumptions for bias adjustment. It combines the validation dataframe
with specific identification of the appropriate columns for bias adjustment,
including: true exposure, true outcome, confounders, misclassified exposure,
misclassified outcome, and selection. The purpose of validation data is to
use an external data source to transport the necessary causal relationships
that are missing in the observed data.
}
\examples{
df <- data_validation(
  data = df_sel_source,
  true_exposure = "X",
  true_outcome = "Y",
  confounders = c("C1", "C2", "C3"),
  selection = "S"
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/multibias_adjust.R
\name{multibias_adjust}
\alias{multibias_adjust}
\title{Simultaneously adjust for multiple biases}
\usage{
multibias_adjust(
  data_observed,
  data_validation = NULL,
  bias_params = NULL,
  bootstrap = FALSE,
  bootstrap_reps = 100,
  level = 0.95
)
}
\arguments{
\item{data_observed}{Object of class \code{data_observed} corresponding to the
data to perform bias analysis on.}

\item{data_validation}{Object of class \code{data_validation} corresponding to
the validation data used to adjust for bias in the observed data. The
validation data should have data for the same variables as in
\code{data_observed}, plus data for the missing variables leading to bias.}

\item{bias_params}{Object of class 'bias_params' corresponding to the
bias parameters used to adjust for bias in the observed data. There must
be parameters corresponding to the bias or biases specified in
\code{data_observed}.}

\item{bootstrap}{Boolean for whether to perform bootstrapping to obtain
the estimate and confidence interval.}

\item{bootstrap_reps}{Integer number of bootstrap samples to run in
bootstrapping.}

\item{level}{Value from 0-1 representing the full range of the confidence
interval. Default is 0.95.}
}
\value{
A list including: the bias-adjusted effect estimate of the exposure
on the outcome, the standard error, and the confidence interval as the
vector: (lower bound, upper bound).
}
\description{
\code{multibias_adjust} returns the exposure-outcome odds ratio and confidence
interval, adjusted for one or more biases.
}
\details{
Bias adjustment can be performed by inputting either a validation dataset or
the necessary bias parameters. Values for the bias parameters
can be applied as fixed values or as single draws from a probability
distribution (ex: \code{rnorm(1, mean = 2, sd = 1)}). The latter has
the advantage of allowing the researcher to capture the uncertainty
in the bias parameter estimates. To incorporate this uncertainty in the
estimate and confidence interval, this function should be run in loop across
bootstrap samples of the dataframe for analysis. The estimate and
confidence interval would then be obtained from the median and quantiles
of the distribution of odds ratio estimates.
}
\examples{
# Adjust for exposure misclassification -------------------------------------
df_observed <- data_observed(
  data = df_em,
  bias = "em",
  exposure = "Xstar",
  outcome = "Y",
  confounders = "C1"
)

# Using validation data
df_validation <- data_validation(
  data = df_em_source,
  true_exposure = "X",
  true_outcome = "Y",
  confounders = "C1",
  misclassified_exposure = "Xstar"
)

multibias_adjust(
  data_observed = df_observed,
  data_validation = df_validation
)

# Using bias_params
bp <- bias_params(coef_list = list(x = c(-2.10, 1.62, 0.63, 0.35)))

multibias_adjust(
  data_observed = df_observed,
  bias_params = bp
)

# Adjust for three biases ---------------------------------------------------
df_observed <- data_observed(
  data = df_uc_om_sel,
  bias = c("uc", "om", "sel"),
  exposure = "X",
  outcome = "Ystar",
  confounders = c("C1", "C2", "C3")
)

# Using validation data
df_validation <- data_validation(
  data = df_uc_om_sel_source,
  true_exposure = "X",
  true_outcome = "Y",
  confounders = c("C1", "C2", "C3", "U"),
  misclassified_outcome = "Ystar",
  selection = "S"
)

multibias_adjust(
  data_observed = df_observed,
  data_validation = df_validation
)

# Using bias_params
bp1 <- bias_params(
  coef_list = list(
    u = c(-0.32, 0.59, 0.69),
    y = c(-2.85, 0.71, 1.63, 0.40, -0.85, 0.22),
    s = c(0.00, 0.74, 0.19, 0.02, -0.06, 0.02)
  )
)

multibias_adjust(
  data_observed = df_observed,
  bias_params = bp1
)

bp2 <- bias_params(
  coef_list = list(
    u1y0 = c(-0.20, 0.62, 0.01, -0.08, 0.10, -0.15),
    u0y1 = c(-3.28, 0.63, 1.65, 0.42, -0.85, 0.26),
    u1y1 = c(-2.70, 1.22, 1.64, 0.32, -0.77, 0.09),
    s = c(0.00, 0.74, 0.19, 0.02, -0.06, 0.02)
  )
)

# with bootstrapping
\dontrun{
multibias_adjust(
  data_observed = df_observed,
  bias_params = bp2,
  bootstrap = TRUE,
  bootstrap_reps = 1000
)
}

}
